# -*- coding: utf-8 -*-

# Copyright © 2012-2024 Roberto Alsina and others.

# Permission is hereby granted, free of charge, to any
# person obtaining a copy of this software and associated
# documentation files (the "Software"), to deal in the
# Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the
# Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice
# shall be included in all copies or substantial portions of
# the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY
# KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
# PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS
# OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR
# OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
# OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

"""Import a WordPress dump."""

import datetime
import io
import json
import os
import re
import sys
from collections import defaultdict
from urllib.parse import urlparse, unquote

import requests
from lxml import etree

from nikola.plugin_categories import Command, CompilerExtension
from nikola import utils, hierarchy_utils
from nikola.nikola import DEFAULT_TRANSLATIONS_PATTERN
from nikola.utils import req_missing
from nikola.plugins.basic_import import ImportMixin, links
from nikola.plugins.command.init import (
    SAMPLE_CONF, prepare_config,
    format_default_translations_config,
    get_default_translations_dict
)

try:
    import html2text
except ImportError:
    html2text = None

try:
    import phpserialize
except ImportError:
    phpserialize = None

LOGGER = utils.get_logger('import_wordpress')


def install_plugin(site, plugin_name, output_dir=None, show_install_notes=False):
    """Install a Nikola plugin."""
    LOGGER.info("Installing plugin '{0}'".format(plugin_name))
    # Get hold of the 'plugin' plugin
    plugin_installer_info = site.plugin_manager.get_plugin_by_name('plugin', 'Command')
    if plugin_installer_info is None:
        LOGGER.error('Internal error: cannot find the "plugin" plugin which is supposed to come with Nikola - it might be disabled in conf.py')
        return False
    plugin_installer = plugin_installer_info.plugin_object
    # Try to install the requested plugin
    options = {}
    for option in plugin_installer.cmd_options:
        options[option['name']] = option['default']
    options['install'] = plugin_name
    options['output_dir'] = output_dir
    options['show_install_notes'] = show_install_notes
    if plugin_installer.execute(options=options) > 0:
        return False
    # Let the plugin manager find newly installed plugins
    old_candidates = set(site.plugin_manager.candidates)
    new_candidates = set(site.plugin_manager.locate_plugins())
    missing_candidates = list(new_candidates - old_candidates)
    new_plugins = site.plugin_manager.load_plugins(missing_candidates)

    # Activate new plugins
    for p in new_plugins:
        site._activate_plugin(p)
        if isinstance(p.plugin_object, CompilerExtension):
            site.compiler_extensions.append(p)
    return True


class CommandImportWordpress(Command, ImportMixin):
    """Import a WordPress dump."""

    name = "import_wordpress"
    needs_config = False
    doc_usage = "[options] wordpress_export_file"
    doc_purpose = "import a WordPress dump"
    cmd_options = ImportMixin.cmd_options + [
        {
            'name': 'exclude_drafts',
            'long': 'no-drafts',
            'short': 'd',
            'default': False,
            'type': bool,
            'help': "Don't import drafts",
        },
        {
            'name': 'exclude_privates',
            'long': 'exclude-privates',
            'default': False,
            'type': bool,
            'help': "Don't import private posts",
        },
        {
            'name': 'include_empty_items',
            'long': 'include-empty-items',
            'default': False,
            'type': bool,
            'help': "Include empty posts and pages",
        },
        {
            'name': 'squash_newlines',
            'long': 'squash-newlines',
            'default': False,
            'type': bool,
            'help': "Shorten multiple newlines in a row to only two newlines",
        },
        {
            'name': 'no_downloads',
            'long': 'no-downloads',
            'default': False,
            'type': bool,
            'help': "Do not try to download files for the import",
        },
        {
            'name': 'download_auth',
            'long': 'download-auth',
            'default': None,
            'type': str,
            'help': "Specify username and password for HTTP authentication (separated by ':')",
        },
        {
            'name': 'separate_qtranslate_content',
            'long': 'qtranslate',
            'default': False,
            'type': bool,
            'help': """Look for translations generated by qtranslate plugin.
WARNING: a default wordpress export won't allow to recover title translations.
For this to be possible consider applying the hack suggested at
https://github.com/qtranslate/qtranslate-xt/issues/199 :

In wp-admin/includes/export.php change
`echo apply_filters( 'the_title_rss', $post->post_title );

to
`echo apply_filters( 'the_title_export', $post->post_title );
"""
        },
        {
            'name': 'translations_pattern',
            'long': 'translations_pattern',
            'default': DEFAULT_TRANSLATIONS_PATTERN,
            'type': str,
            'help': "The pattern for translation files names",
        },
        {
            'name': 'export_categories_as_categories',
            'long': 'export-categories-as-categories',
            'default': False,
            'type': bool,
            'help': "Export categories as categories, instead of treating them as tags",
        },
        {
            'name': 'export_comments',
            'long': 'export-comments',
            'default': False,
            'type': bool,
            'help': "Export comments as .wpcomment files",
        },
        {
            'name': 'html2text',
            'long': 'html2text',
            'default': False,
            'type': bool,
            'help': "Uses html2text (needs to be installed with pip) to transform WordPress posts to MarkDown during import",
        },
        {
            'name': 'transform_to_markdown',
            'long': 'transform-to-markdown',
            'default': False,
            'type': bool,
            'help': "Uses WordPress page compiler to transform WordPress posts to HTML and then use html2text to transform them to MarkDown during import",
        },
        {
            'name': 'transform_to_html',
            'long': 'transform-to-html',
            'default': False,
            'type': bool,
            'help': "Uses WordPress page compiler to transform WordPress posts directly to HTML during import",
        },
        {
            'name': 'use_wordpress_compiler',
            'long': 'use-wordpress-compiler',
            'default': False,
            'type': bool,
            'help': "Instead of converting posts to markdown, leave them as is and use the WordPress page compiler",
        },
        {
            'name': 'install_wordpress_compiler',
            'long': 'install-wordpress-compiler',
            'default': False,
            'type': bool,
            'help': "Automatically installs the WordPress page compiler (either locally or in the new site) if required by other options.\nWarning: the compiler is GPL software!",
        },
        {
            'name': 'tag_sanitizing_strategy',
            'long': 'tag-sanitizing-strategy',
            'default': 'first',
            'help': 'lower: Convert all tag and category names to lower case\nfirst: Keep first spelling of tag or category name',
        },
        {
            'name': 'one_file',
            'long': 'one-file',
            'default': False,
            'type': bool,
            'help': "Save imported posts in the more modern one-file format.",
        },
    ]
    all_tags = set([])

    def _get_compiler(self):
        """Return whatever compiler we will use."""
        self._find_wordpress_compiler()
        if self.wordpress_page_compiler is not None:
            return self.wordpress_page_compiler
        plugin_info = self.site.plugin_manager.get_plugin_by_name('markdown', 'PageCompiler')
        if plugin_info is not None:
            if not plugin_info.is_activated:
                plugin_info.plugin_object.set_site(self.site)
            return plugin_info.plugin_object
        else:
            LOGGER.error("Can't find markdown post compiler.")

    def _find_wordpress_compiler(self):
        """Find WordPress compiler plugin."""
        if self.wordpress_page_compiler is not None:
            return

        plugin_info = self.site.plugin_manager.get_plugin_by_name('wordpress', 'PageCompiler')
        if plugin_info is None:
            candidates = self.site.plugin_manager.locate_plugins()
            wordpress_candidates = [c for c in candidates if c.name == "wordpress" and c.category == "PageCompiler"]
            if wordpress_candidates:
                new_plugins = self.site.plugin_manager.load_plugins(wordpress_candidates)
                for p in new_plugins:
                    self.site._activate_plugin(p)
                    self.wordpress_page_compiler = p.plugin_object
        else:
            self.wordpress_page_compiler = plugin_info.plugin_object

    def _read_options(self, options, args):
        """Read command-line options."""
        options['filename'] = args.pop(0)

        if args and ('output_folder' not in args or
                     options['output_folder'] == 'new_site'):
            options['output_folder'] = args.pop(0)

        if args:
            LOGGER.warning('You specified additional arguments ({0}). Please consider '
                           'putting these arguments before the filename if you '
                           'are running into problems.'.format(args))

        self.onefile = options.get('one_file', False)

        self.import_into_existing_site = False
        self.url_map = {}
        self.timezone = None

        self.wordpress_export_file = options['filename']
        self.squash_newlines = options.get('squash_newlines', False)
        self.output_folder = options.get('output_folder', 'new_site')

        self.exclude_drafts = options.get('exclude_drafts', False)
        self.exclude_privates = options.get('exclude_privates', False)
        self.no_downloads = options.get('no_downloads', False)
        self.import_empty_items = options.get('include_empty_items', False)

        self.export_categories_as_categories = options.get('export_categories_as_categories', False)
        self.export_comments = options.get('export_comments', False)

        self.html2text = options.get('html2text', False)
        self.transform_to_markdown = options.get('transform_to_markdown', False)

        self.transform_to_html = options.get('transform_to_html', False)
        self.use_wordpress_compiler = options.get('use_wordpress_compiler', False)
        self.install_wordpress_compiler = options.get('install_wordpress_compiler', False)
        self.wordpress_page_compiler = None

        self.tag_saniziting_strategy = options.get('tag_saniziting_strategy', 'first')

        self.auth = None
        if options.get('download_auth') is not None:
            username_password = options.get('download_auth')
            self.auth = tuple(username_password.split(':', 1))
            if len(self.auth) < 2:
                LOGGER.error("Please specify HTTP authentication credentials in the form username:password.")
                return False

        self.separate_qtranslate_content = options.get('separate_qtranslate_content')
        self.translations_pattern = options.get('translations_pattern')

        count = (1 if self.html2text else 0) + (1 if self.transform_to_html else 0) + (1 if self.transform_to_markdown else 0)
        if count > 1:
            LOGGER.error("You can use at most one of the options --html2text, --transform-to-html and --transform-to-markdown.")
            return False
        if (self.html2text or self.transform_to_html or self.transform_to_markdown) and self.use_wordpress_compiler:
            LOGGER.warning("It does not make sense to combine --use-wordpress-compiler with any of --html2text, --transform-to-html and --transform-to-markdown, as the latter convert all posts to HTML and the first option then affects zero posts.")

        if (self.html2text or self.transform_to_markdown) and not html2text:
            LOGGER.error("You need to install html2text via 'pip install html2text' before you can use the --html2text and --transform-to-markdown options.")
            return False

        if self.transform_to_html or self.transform_to_markdown:
            self._find_wordpress_compiler()
            if not self.wordpress_page_compiler and self.install_wordpress_compiler:
                if not install_plugin(self.site, 'wordpress_compiler', output_dir='plugins'):  # local install
                    return False
                self._find_wordpress_compiler()
            if not self.wordpress_page_compiler:
                LOGGER.error("To compile WordPress posts to HTML, the WordPress post compiler is needed. You can install it via:")
                LOGGER.error("    nikola plugin -i wordpress_compiler")
                LOGGER.error("Please note that the WordPress post compiler is licensed under the GPL v2.")
                return False

        return True

    def _prepare(self, channel):
        """Prepare context and category hierarchy."""
        self.context = self.populate_context(channel)
        self.base_dir = urlparse(self.context['BASE_URL']).path

        if self.export_categories_as_categories:
            wordpress_namespace = channel.nsmap['wp']
            cat_map = dict()
            for cat in channel.findall('{{{0}}}category'.format(wordpress_namespace)):
                # cat_id = get_text_tag(cat, '{{{0}}}term_id'.format(wordpress_namespace), None)
                cat_slug = get_text_tag(cat, '{{{0}}}category_nicename'.format(wordpress_namespace), None)
                cat_parent_slug = get_text_tag(cat, '{{{0}}}category_parent'.format(wordpress_namespace), None)
                cat_name = utils.html_unescape(get_text_tag(cat, '{{{0}}}cat_name'.format(wordpress_namespace), None))
                cat_path = [cat_name]
                if cat_parent_slug in cat_map:
                    cat_path = cat_map[cat_parent_slug] + cat_path
                cat_map[cat_slug] = cat_path
            self._category_paths = dict()
            for cat, path in cat_map.items():
                self._category_paths[cat] = hierarchy_utils.join_hierarchical_category_path(path)

    def _execute(self, options={}, args=[]):
        """Import a WordPress blog from an export file into a Nikola site."""
        if not args:
            print(self.help())
            return False

        if not self._read_options(options, args):
            return False

        # A place holder where extra language (if detected) will be stored
        self.extra_languages = set()

        if not self.no_downloads:
            def show_info_about_mising_module(modulename):
                LOGGER.error(
                    'To use the "{commandname}" command, you have to install '
                    'the "{package}" package or supply the "--no-downloads" '
                    'option.'.format(
                        commandname=self.name,
                        package=modulename)
                )

            if phpserialize is None:
                req_missing(['phpserialize'], 'import WordPress dumps without --no-downloads')

        export_file_preprocessor = modernize_qtranslate_tags if self.separate_qtranslate_content else None
        channel = self.get_channel_from_file(self.wordpress_export_file, export_file_preprocessor)
        self._prepare(channel)
        conf_template = self.generate_base_site()

        self.import_posts(channel)
        self.context['TRANSLATIONS'] = format_default_translations_config(
            self.extra_languages)
        self.context['REDIRECTIONS'] = self.configure_redirections(
            self.url_map, self.base_dir)
        if self.timezone:
            self.context['TIMEZONE'] = self.timezone
        if self.export_categories_as_categories:
            self.context['CATEGORY_ALLOW_HIERARCHIES'] = True
            self.context['CATEGORY_OUTPUT_FLAT_HIERARCHY'] = True

        # Add tag redirects
        for tag in self.all_tags:
            try:
                if isinstance(tag, bytes):
                    tag_str = tag.decode('utf8', 'replace')
                else:
                    tag_str = tag
            except AttributeError:
                tag_str = tag
            tag = utils.slugify(tag_str, self.lang)
            src_url = '{}tag/{}'.format(self.context['SITE_URL'], tag)
            dst_url = self.site.link('tag', tag)
            if src_url != dst_url:
                self.url_map[src_url] = dst_url

        self.write_urlmap_csv(
            os.path.join(self.output_folder, 'url_map.csv'), self.url_map)
        rendered_template = conf_template.render(**prepare_config(self.context))
        self.write_configuration(self.get_configuration_output_path(),
                                 rendered_template)

        if self.use_wordpress_compiler:
            if self.install_wordpress_compiler:
                if not install_plugin(self.site, 'wordpress_compiler', output_dir=os.path.join(self.output_folder, 'plugins')):
                    return False
            else:
                LOGGER.warning("Make sure to install the WordPress page compiler via")
                LOGGER.warning("    nikola plugin -i wordpress_compiler")
                LOGGER.warning("in your imported blog's folder ({0}), if you haven't installed it system-wide or user-wide. Otherwise, your newly imported blog won't compile.".format(self.output_folder))

    @classmethod
    def read_xml_file(cls, filename):
        """Read XML file into memory."""
        xml = []

        with open(filename, 'rb') as fd:
            for line in fd:
                # These explode etree and are useless
                if b'<atom:link rel=' in line:
                    continue
                xml.append(line)
        return b''.join(xml)

    @classmethod
    def get_channel_from_file(cls, filename, xml_preprocessor=None):
        """Get channel from XML file.

        An optional 'xml_preprocessor' allows to modify the xml
        (typically to deal with variations in tags injected by some WP plugin)
        """
        xml_string = cls.read_xml_file(filename)
        if xml_preprocessor:
            xml_string = xml_preprocessor(xml_string)
        tree = etree.fromstring(xml_string)
        channel = tree.find('channel')
        return channel

    def populate_context(self, channel):
        """Populate context with config for the site."""
        wordpress_namespace = channel.nsmap['wp']

        context = SAMPLE_CONF.copy()
        self.lang = get_text_tag(channel, 'language', 'en')[:2]
        context['DEFAULT_LANG'] = self.lang
        # If user  has specified a custom pattern for translation files we
        # need to fix the config
        context['TRANSLATIONS_PATTERN'] = self.translations_pattern

        context['BLOG_TITLE'] = get_text_tag(channel, 'title',
                                             'PUT TITLE HERE')
        context['BLOG_DESCRIPTION'] = get_text_tag(
            channel, 'description', 'PUT DESCRIPTION HERE')
        context['BASE_URL'] = get_text_tag(channel, 'link', '#')
        if not context['BASE_URL']:
            base_site_url = channel.find('{{{0}}}author'.format(wordpress_namespace))
            context['BASE_URL'] = get_text_tag(base_site_url,
                                               None,
                                               "http://foo.com/")
        if not context['BASE_URL'].endswith('/'):
            context['BASE_URL'] += '/'
        context['SITE_URL'] = context['BASE_URL']

        author = channel.find('{{{0}}}author'.format(wordpress_namespace))
        context['BLOG_EMAIL'] = get_text_tag(
            author,
            '{{{0}}}author_email'.format(wordpress_namespace),
            "joe@example.com")
        context['BLOG_AUTHOR'] = get_text_tag(
            author,
            '{{{0}}}author_display_name'.format(wordpress_namespace),
            "Joe Example")
        extensions = ['rst', 'txt', 'md', 'html']
        if self.use_wordpress_compiler:
            extensions.append('wp')
        POSTS = '(\n'
        PAGES = '(\n'
        for extension in extensions:
            POSTS += '    ("posts/*.{0}", "posts", "post.tmpl"),\n'.format(extension)
            PAGES += '    ("pages/*.{0}", "pages", "page.tmpl"),\n'.format(extension)
        POSTS += ')\n'
        PAGES += ')\n'
        context['POSTS'] = POSTS
        context['PAGES'] = PAGES
        COMPILERS = '{\n'
        COMPILERS += '''    "rest": ['.txt', '.rst'],''' + '\n'
        COMPILERS += '''    "markdown": ['.md', '.mdown', '.markdown'],''' + '\n'
        COMPILERS += '''    "html": ['.html', '.htm'],''' + '\n'
        if self.use_wordpress_compiler:
            COMPILERS += '''    "wordpress": ['.wp'],''' + '\n'
        COMPILERS += '}'
        context['COMPILERS'] = COMPILERS

        return context

    def download_url_content_to_file(self, url, dst_path):
        """Download some content (attachments) to a file."""
        try:
            request = requests.get(url, auth=self.auth)
            if request.status_code >= 400:
                LOGGER.warning("Downloading {0} to {1} failed with HTTP status code {2}".format(url, dst_path, request.status_code))
                return
            with open(dst_path, 'wb+') as fd:
                fd.write(request.content)
        except requests.exceptions.ConnectionError as err:
            LOGGER.warning("Downloading {0} to {1} failed: {2}".format(url, dst_path, err))

    def import_attachment(self, item, wordpress_namespace):
        """Import an attachment to the site."""
        # Download main image
        url = get_text_tag(
            item, '{{{0}}}attachment_url'.format(wordpress_namespace), 'foo')
        link = get_text_tag(item, '{{{0}}}link'.format(wordpress_namespace),
                            'foo')
        path = urlparse(url).path
        dst_path = os.path.join(*([self.output_folder, 'files'] + list(path.split('/'))))
        if self.no_downloads:
            LOGGER.info("Skipping downloading {0} => {1}".format(url, dst_path))
        else:
            dst_dir = os.path.dirname(dst_path)
            utils.makedirs(dst_dir)
            LOGGER.info("Downloading {0} => {1}".format(url, dst_path))
            self.download_url_content_to_file(url, dst_path)
        dst_url = '/'.join(dst_path.split(os.sep)[2:])
        links[link] = '/' + dst_url
        links[url] = '/' + dst_url

        files = [path]
        files_meta = [{}]

        additional_metadata = item.findall('{{{0}}}postmeta'.format(wordpress_namespace))
        if phpserialize and additional_metadata:
            source_path = os.path.dirname(url)
            for element in additional_metadata:
                meta_key = element.find('{{{0}}}meta_key'.format(wordpress_namespace))
                if meta_key is not None and meta_key.text == '_wp_attachment_metadata':
                    meta_value = element.find('{{{0}}}meta_value'.format(wordpress_namespace))

                    if meta_value is None:
                        continue

                    # Someone from Wordpress thought it was a good idea
                    # serialize PHP objects into that metadata field. Given
                    # that the export should give you the power to insert
                    # your blogging into another site or system its not.
                    # Why don't they just use JSON?
                    metadata = phpserialize.loads(meta_value.text.encode('utf-8'))

                    meta_key = b'image_meta'
                    size_key = b'sizes'
                    file_key = b'file'
                    width_key = b'width'
                    height_key = b'height'

                    # Extract metadata
                    if width_key in metadata and height_key in metadata:
                        files_meta[0]['width'] = int(metadata[width_key])
                        files_meta[0]['height'] = int(metadata[height_key])

                    if meta_key in metadata:
                        image_meta = metadata[meta_key]
                        if not image_meta:
                            continue
                        dst_meta = {}

                        def add(our_key, wp_key, is_int=False, ignore_zero=False, is_float=False):
                            if wp_key in image_meta:
                                value = image_meta[wp_key]
                                if value in ("", b"") and (is_int or is_float):
                                    if ignore_zero:
                                        return
                                    else:
                                        dst_meta[our_key] = 0
                                if is_int:
                                    value = int(value)
                                    if ignore_zero and value == 0:
                                        return
                                elif is_float:
                                    # in some locales (like fr) and for old posts there may be a comma here.
                                    if isinstance(value, bytes):
                                        value = value.replace(b",", b".")
                                    value = float(value)
                                    if ignore_zero and value == 0:
                                        return
                                else:
                                    value = value.decode('utf-8')  # assume UTF-8
                                    if value == '':  # skip empty values
                                        return
                                dst_meta[our_key] = value

                        add('aperture', b'aperture', is_float=True, ignore_zero=True)
                        add('credit', b'credit')
                        add('camera', b'camera')
                        add('caption', b'caption')
                        add('created_timestamp', b'created_timestamp', is_float=True, ignore_zero=True)
                        add('copyright', b'copyright')
                        add('focal_length', b'focal_length', is_float=True, ignore_zero=True)
                        add('iso', b'iso', is_float=True, ignore_zero=True)
                        add('shutter_speed', b'shutter_speed', ignore_zero=True, is_float=True)
                        add('title', b'title')

                        if len(dst_meta) > 0:
                            files_meta[0]['meta'] = dst_meta

                    # Find other sizes of image
                    if size_key not in metadata:
                        continue

                    for size in metadata[size_key]:
                        filename = metadata[size_key][size][file_key]
                        url = '/'.join([source_path, filename.decode('utf-8')])

                        # Construct metadata
                        meta = {}
                        meta['size'] = size.decode('utf-8')
                        if width_key in metadata[size_key][size] and height_key in metadata[size_key][size]:
                            meta['width'] = int(metadata[size_key][size][width_key])
                            meta['height'] = int(metadata[size_key][size][height_key])

                        path = urlparse(url).path
                        dst_path = os.path.join(*([self.output_folder, 'files'] + list(path.split('/'))))
                        if self.no_downloads:
                            LOGGER.info("Skipping downloading {0} => {1}".format(url, dst_path))
                        else:
                            dst_dir = os.path.dirname(dst_path)
                            utils.makedirs(dst_dir)
                            LOGGER.info("Downloading {0} => {1}".format(url, dst_path))
                            self.download_url_content_to_file(url, dst_path)
                        dst_url = '/'.join(dst_path.split(os.sep)[2:])
                        links[url] = '/' + dst_url

                        files.append(path)
                        files_meta.append(meta)

        # Prepare result
        result = {}
        result['files'] = files
        result['files_meta'] = files_meta

        # Prepare extraction of more information
        dc_namespace = item.nsmap['dc']
        content_namespace = item.nsmap['content']
        excerpt_namespace = item.nsmap['excerpt']

        def add(result_key, key, namespace=None, filter=None, store_empty=False):
            if namespace is not None:
                value = get_text_tag(item, '{{{0}}}{1}'.format(namespace, key), None)
            else:
                value = get_text_tag(item, key, None)
            if value is not None:
                if filter:
                    value = filter(value)
                if value or store_empty:
                    result[result_key] = value

        add('title', 'title')
        add('date_utc', 'post_date_gmt', namespace=wordpress_namespace)
        add('wordpress_user_name', 'creator', namespace=dc_namespace)
        add('content', 'encoded', namespace=content_namespace)
        add('excerpt', 'encoded', namespace=excerpt_namespace)
        add('description', 'description')

        return result

    code_re1 = re.compile(r'\[code.* lang.*?="(.*?)?".*\](.*?)\[/code\]', re.DOTALL | re.MULTILINE)
    code_re2 = re.compile(r'\[sourcecode.* lang.*?="(.*?)?".*\](.*?)\[/sourcecode\]', re.DOTALL | re.MULTILINE)
    code_re3 = re.compile(r'\[code.*?\](.*?)\[/code\]', re.DOTALL | re.MULTILINE)
    code_re4 = re.compile(r'\[sourcecode.*?\](.*?)\[/sourcecode\]', re.DOTALL | re.MULTILINE)

    def transform_code(self, content):
        """Transform code blocks."""
        # https://en.support.wordpress.com/code/posting-source-code/. There are
        # a ton of things not supported here. We only do a basic [code
        # lang="x"] -> ```x translation, and remove quoted html entities (<,
        # >, &, and ").
        def replacement(m, c=content):
            if len(m.groups()) == 1:
                language = ''
                code = m.group(0)
            else:
                language = m.group(1) or ''
                code = m.group(2)
            code = code.replace('&amp;', '&')
            code = code.replace('&gt;', '>')
            code = code.replace('&lt;', '<')
            code = code.replace('&quot;', '"')
            return '```{language}\n{code}\n```'.format(language=language, code=code)

        content = self.code_re1.sub(replacement, content)
        content = self.code_re2.sub(replacement, content)
        content = self.code_re3.sub(replacement, content)
        content = self.code_re4.sub(replacement, content)
        return content

    @staticmethod
    def transform_caption(content, use_html=False):
        """Transform captions."""
        new_caption = re.sub(r'\[/caption\]', '</h1>' if use_html else '', content)
        new_caption = re.sub(r'\[caption.*\]', '<h1>' if use_html else '', new_caption)

        return new_caption

    def transform_multiple_newlines(self, content):
        """Replace multiple newlines with only two."""
        if self.squash_newlines:
            return re.sub(r'\n{3,}', r'\n\n', content)
        else:
            return content

    def transform_content(self, content, post_format, attachments):
        """Transform content into appropriate format."""
        if post_format == 'wp':
            if self.transform_to_html:
                additional_data = {}
                if attachments is not None:
                    additional_data['attachments'] = attachments
                try:
                    content = self.wordpress_page_compiler.compile_to_string(content, additional_data=additional_data)
                except TypeError:  # old versions of the plugin don't support the additional argument
                    content = self.wordpress_page_compiler.compile_to_string(content)
                return content, 'html', True
            elif self.transform_to_markdown:
                # First convert to HTML with WordPress plugin
                additional_data = {}
                if attachments is not None:
                    additional_data['attachments'] = attachments
                try:
                    content = self.wordpress_page_compiler.compile_to_string(content, additional_data=additional_data)
                except TypeError:  # old versions of the plugin don't support the additional argument
                    content = self.wordpress_page_compiler.compile_to_string(content)
                # Now convert to MarkDown with html2text
                h = html2text.HTML2Text()
                content = h.handle(content)
                return content, 'md', False
            elif self.html2text:
                # TODO: what to do with [code] blocks?
                # content = self.transform_code(content)
                content = self.transform_caption(content, use_html=True)
                h = html2text.HTML2Text()
                content = h.handle(content)
                return content, 'md', False
            elif self.use_wordpress_compiler:
                return content, 'wp', False
            else:
                content = self.transform_code(content)
                content = self.transform_caption(content)
                content = self.transform_multiple_newlines(content)
                return content, 'md', True
        elif post_format == 'markdown':
            return content, 'md', True
        elif post_format == 'none':
            return content, 'html', True
        else:
            return None

    def _extract_comment(self, comment, wordpress_namespace):
        """Extract comment from dump."""
        id = int(get_text_tag(comment, "{{{0}}}comment_id".format(wordpress_namespace), None))
        author = get_text_tag(comment, "{{{0}}}comment_author".format(wordpress_namespace), None)
        author_email = get_text_tag(comment, "{{{0}}}comment_author_email".format(wordpress_namespace), None)
        author_url = get_text_tag(comment, "{{{0}}}comment_author_url".format(wordpress_namespace), None)
        author_IP = get_text_tag(comment, "{{{0}}}comment_author_IP".format(wordpress_namespace), None)
        # date = get_text_tag(comment, "{{{0}}}comment_date".format(wordpress_namespace), None)
        date_gmt = get_text_tag(comment, "{{{0}}}comment_date_gmt".format(wordpress_namespace), None)
        content = get_text_tag(comment, "{{{0}}}comment_content".format(wordpress_namespace), None)
        approved = get_text_tag(comment, "{{{0}}}comment_approved".format(wordpress_namespace), '0')
        if approved == '0':
            approved = 'hold'
        elif approved == '1':
            approved = 'approved'
        elif approved == 'spam' or approved == 'trash':
            pass
        else:
            LOGGER.warning("Unknown comment approved status: {0}".format(approved))
        parent = int(get_text_tag(comment, "{{{0}}}comment_parent".format(wordpress_namespace), 0))
        if parent == 0:
            parent = None
        user_id = int(get_text_tag(comment, "{{{0}}}comment_user_id".format(wordpress_namespace), 0))
        if user_id == 0:
            user_id = None

        if approved == 'trash' or approved == 'spam':
            return None

        return {"id": id, "status": str(approved), "approved": approved == "approved",
                "author": author, "email": author_email, "url": author_url, "ip": author_IP,
                "date": date_gmt, "content": content, "parent": parent, "user_id": user_id}

    def _write_comment(self, filename, comment):
        """Write comment to file."""
        def write_header_line(fd, header_field, header_content):
            """Write comment header line."""
            if header_content is None:
                return
            header_content = str(header_content).replace('\n', ' ')
            line = '.. ' + header_field + ': ' + header_content + '\n'
            fd.write(line.encode('utf8'))

        with open(filename, "wb+") as fd:
            write_header_line(fd, "id", comment["id"])
            write_header_line(fd, "status", comment["status"])
            write_header_line(fd, "approved", comment["approved"])
            write_header_line(fd, "author", comment["author"])
            write_header_line(fd, "author_email", comment["email"])
            write_header_line(fd, "author_url", comment["url"])
            write_header_line(fd, "author_IP", comment["ip"])
            write_header_line(fd, "date_utc", comment["date"])
            write_header_line(fd, "parent_id", comment["parent"])
            write_header_line(fd, "wordpress_user_id", comment["user_id"])
            fd.write(('\n' + comment['content']).encode('utf8'))

    def _create_meta_and_content_filenames(self, slug, extension, lang, default_language, translations_config):
        out_meta_filename = slug + '.meta'
        out_content_filename = slug + '.' + extension
        if lang and lang != default_language:
            out_meta_filename = utils.get_translation_candidate(translations_config,
                                                                out_meta_filename, lang)
            out_content_filename = utils.get_translation_candidate(translations_config,
                                                                   out_content_filename, lang)
        return out_meta_filename, out_content_filename

    def _create_metadata(self, status, excerpt, tags, categories, post_name=None):
        """Create post metadata."""
        other_meta = {'wp-status': status}
        if excerpt is not None:
            other_meta['excerpt'] = excerpt
        if self.export_categories_as_categories:
            cats = []
            for text in categories:
                if text in self._category_paths:
                    cats.append(self._category_paths[text])
                else:
                    cats.append(hierarchy_utils.join_hierarchical_category_path([utils.html_unescape(text)]))
            other_meta['categories'] = ','.join(cats)
            if len(cats) > 0:
                other_meta['category'] = cats[0]
                if len(cats) > 1:
                    LOGGER.warning(('Post "{0}" has more than one category! ' +
                                    'Will only use the first one.').format(post_name))
            tags_cats = [utils.html_unescape(tag) for tag in tags]
        else:
            tags_cats = [utils.html_unescape(tag) for tag in tags + categories]
        return tags_cats, other_meta

    _tag_sanitize_map = {True: {}, False: {}}

    def _sanitize(self, tag, is_category):
        if self.tag_saniziting_strategy == 'lower':
            return tag.lower()
        if tag.lower() not in self._tag_sanitize_map[is_category]:
            self._tag_sanitize_map[is_category][tag.lower()] = [tag]
            return tag
        previous = self._tag_sanitize_map[is_category][tag.lower()]
        if self.tag_saniziting_strategy == 'first':
            if tag != previous[0]:
                LOGGER.warning("Changing spelling of {0} name '{1}' to {2}.".format('category' if is_category else 'tag', tag, previous[0]))
            return previous[0]
        else:
            LOGGER.error("Unknown tag sanitizing strategy '{0}'!".format(self.tag_saniziting_strategy))
            sys.exit(1)
        return tag

    def import_postpage_item(self, item, wordpress_namespace, out_folder=None, attachments=None):
        """Take an item from the feed and creates a post file."""
        if out_folder is None:
            out_folder = 'posts'

        title = get_text_tag(item, 'title', 'NO TITLE')

        # titles can have line breaks in them, particularly when they are
        # created by third-party tools that post to Wordpress.
        # Handle windows-style and unix-style line endings.
        title = title.replace('\r\n', ' ').replace('\n', ' ')

        # link is something like http://foo.com/2012/09/01/hello-world/
        # So, take the path, utils.slugify it, and that's our slug
        link = get_text_tag(item, 'link', None)
        parsed = urlparse(link)
        path = unquote(parsed.path.strip('/'))

        try:
            if isinstance(path, bytes):
                path = path.decode('utf8', 'replace')
            else:
                path = path
        except AttributeError:
            pass

        # Cut out the base directory.
        if path.startswith(self.base_dir.strip('/')):
            path = path.replace(self.base_dir.strip('/'), '', 1)

        pathlist = path.split('/')
        if parsed.query:  # if there are no nice URLs and query strings are used
            out_folder = os.path.join(*([out_folder] + pathlist))
            slug = get_text_tag(
                item, '{{{0}}}post_name'.format(wordpress_namespace), None)
            if not slug:  # it *may* happen
                slug = get_text_tag(
                    item, '{{{0}}}post_id'.format(wordpress_namespace), None)
            if not slug:  # should never happen
                LOGGER.error("Error converting post:", title)
                return False
        else:
            if len(pathlist) > 1:
                out_folder = os.path.join(*([out_folder] + pathlist[:-1]))
            slug = utils.slugify(pathlist[-1], self.lang)

        description = get_text_tag(item, 'description', '')
        post_date = get_text_tag(
            item, '{{{0}}}post_date'.format(wordpress_namespace), None)
        try:
            dt = utils.to_datetime(post_date)
        except ValueError:
            dt = datetime.datetime(1970, 1, 1, 0, 0, 0)
            LOGGER.error('Malformed date "{0}" in "{1}" [{2}], assuming 1970-01-01 00:00:00 instead.'.format(post_date, title, slug))
            post_date = dt.strftime('%Y-%m-%d %H:%M:%S')

        if dt.tzinfo and self.timezone is None:
            self.timezone = utils.get_tzname(dt)
        status = get_text_tag(
            item, '{{{0}}}status'.format(wordpress_namespace), 'publish')
        content = get_text_tag(
            item, '{http://purl.org/rss/1.0/modules/content/}encoded', '')
        excerpt = get_text_tag(
            item, '{http://wordpress.org/export/1.2/excerpt/}encoded', None)

        if excerpt is not None:
            if len(excerpt) == 0:
                excerpt = None

        tags = []
        categories = []
        post_status = 'published'
        has_math = "no"
        if status == 'trash':
            LOGGER.warning('Trashed post "{0}" will not be imported.'.format(title))
            return False
        elif status == 'private':
            is_draft = False
            is_private = True
            post_status = 'private'
        elif status != 'publish':
            is_draft = True
            is_private = False
            post_status = 'draft'
        else:
            is_draft = False
            is_private = False

        for tag in item.findall('category'):
            text = tag.text
            type = 'category'
            if 'domain' in tag.attrib:
                type = tag.attrib['domain']
            if text == 'Uncategorized' and type == 'category':
                continue
            if type == 'category':
                categories.append(text)
            else:
                tags.append(text)

        if '$latex' in content:
            has_math = "yes"

        for i, cat in enumerate(categories[:]):
            cat = self._sanitize(cat, True)
            categories[i] = cat
            self.all_tags.add(cat)

        for i, tag in enumerate(tags[:]):
            tag = self._sanitize(tag, False)
            tags[i] = tag
            self.all_tags.add(tag)

        # Find post format if it's there
        post_format = 'wp'
        format_tag = [x for x in item.findall('*//{%s}meta_key' % wordpress_namespace) if x.text == '_tc_post_format']
        if format_tag:
            post_format = format_tag[0].getparent().find('{%s}meta_value' % wordpress_namespace).text
            if post_format == 'wpautop':
                post_format = 'wp'

        if is_draft and self.exclude_drafts:
            LOGGER.warning('Draft "{0}" will not be imported.'.format(title))
            return False
        elif is_private and self.exclude_privates:
            LOGGER.warning('Private post "{0}" will not be imported.'.format(title))
            return False
        elif content.strip() or self.import_empty_items:
            # If no content is found, no files are written.
            self.url_map[link] = (self.context['SITE_URL'] +
                                  out_folder.rstrip('/') + '/' + slug +
                                  '.html').replace(os.sep, '/')
            default_language = self.context["DEFAULT_LANG"]
            if self.separate_qtranslate_content:
                content_translations = separate_qtranslate_tagged_langs(content)
                title_translations = separate_qtranslate_tagged_langs(title)
            else:
                content_translations = {"": content}
                title_translations = {"": title}
            # in case of mistmatch between the languages found in the title and in the content
            default_title = title_translations.get(default_language, title)
            extra_languages = [lang for lang in content_translations.keys() if lang not in ("", default_language)]
            for extra_lang in extra_languages:
                self.extra_languages.add(extra_lang)
            translations_dict = get_default_translations_dict(default_language, extra_languages)
            current_translations_config = {
                "DEFAULT_LANG": default_language,
                "TRANSLATIONS": translations_dict,
                "TRANSLATIONS_PATTERN": self.context["TRANSLATIONS_PATTERN"]
            }
            for lang, content in content_translations.items():
                try:
                    content, extension, rewrite_html = self.transform_content(content, post_format, attachments)
                except Exception:
                    LOGGER.error(('Cannot interpret post "{0}" (language {1}) with post ' +
                                  'format {2}!').format(os.path.join(out_folder, slug), lang, post_format))
                    return False

                out_meta_filename, out_content_filename = self._create_meta_and_content_filenames(
                    slug, extension, lang, default_language, current_translations_config)

                tags, other_meta = self._create_metadata(status, excerpt, tags, categories,
                                                         post_name=os.path.join(out_folder, slug))
                current_title = title_translations.get(lang, default_title)
                meta = {
                    "title": current_title,
                    "slug": slug,
                    "date": post_date,
                    "description": description,
                    "tags": ','.join(tags),
                    "status": post_status,
                    "has_math": has_math,
                }
                meta.update(other_meta)
                if self.onefile:
                    self.write_post(
                        os.path.join(self.output_folder,
                                     out_folder, out_content_filename),
                        content,
                        meta,
                        self._get_compiler(),
                        rewrite_html)
                else:
                    self.write_metadata(os.path.join(self.output_folder, out_folder,
                                                     out_meta_filename),
                                        current_title, slug, post_date, description, tags, **other_meta)
                    self.write_content(
                        os.path.join(self.output_folder,
                                     out_folder, out_content_filename),
                        content,
                        rewrite_html)

            if self.export_comments:
                comments = []
                for tag in item.findall('{{{0}}}comment'.format(wordpress_namespace)):
                    comment = self._extract_comment(tag, wordpress_namespace)
                    if comment is not None:
                        comments.append(comment)

                for comment in comments:
                    comment_filename = "{0}.{1}.wpcomment".format(slug, comment['id'])
                    self._write_comment(os.path.join(self.output_folder, out_folder, comment_filename), comment)

            return (out_folder, slug)
        else:
            LOGGER.warning(('Not going to import "{0}" because it seems to contain'
                            ' no content.').format(title))
            return False

    def _extract_item_info(self, item):
        """Extract information about an item."""
        # The namespace usually is something like:
        # http://wordpress.org/export/1.2/
        wordpress_namespace = item.nsmap['wp']
        post_type = get_text_tag(
            item, '{{{0}}}post_type'.format(wordpress_namespace), 'post')
        post_id = int(get_text_tag(
            item, '{{{0}}}post_id'.format(wordpress_namespace), "0"))
        parent_id = get_text_tag(
            item, '{{{0}}}post_parent'.format(wordpress_namespace), None)
        return wordpress_namespace, post_type, post_id, parent_id

    def process_item_if_attachment(self, item):
        """Process attachments."""
        wordpress_namespace, post_type, post_id, parent_id = self._extract_item_info(item)

        if post_type == 'attachment':
            data = self.import_attachment(item, wordpress_namespace)
            # If parent was found, store relation with imported files
            if parent_id is not None and int(parent_id) != 0:
                self.attachments[int(parent_id)][post_id] = data
            else:
                LOGGER.warning("Attachment #{0} ({1}) has no parent!".format(post_id, data['files']))

    def write_attachments_info(self, path, attachments):
        """Write attachments info file."""
        with io.open(path, "wb") as file:
            file.write(json.dumps(attachments).encode('utf-8'))

    def process_item_if_post_or_page(self, item):
        """Process posts and pages."""
        wordpress_namespace, post_type, post_id, parent_id = self._extract_item_info(item)

        if post_type != 'attachment':
            # Get attachments for post
            attachments = self.attachments.pop(post_id, None)
            # Import item
            if post_type == 'post':
                out_folder_slug = self.import_postpage_item(item, wordpress_namespace, 'posts', attachments)
            else:
                out_folder_slug = self.import_postpage_item(item, wordpress_namespace, 'pages', attachments)
            # Process attachment data
            if attachments is not None:
                # If post was exported, store data
                if out_folder_slug:
                    destination = os.path.join(self.output_folder, out_folder_slug[0],
                                               out_folder_slug[1] + ".attachments.json")
                    self.write_attachments_info(destination, attachments)

    def import_posts(self, channel):
        """Import posts into the site."""
        self.attachments = defaultdict(dict)
        # First process attachments
        for item in channel.findall('item'):
            self.process_item_if_attachment(item)
        # Next process posts
        for item in channel.findall('item'):
            self.process_item_if_post_or_page(item)
        # Assign attachments to posts
        for post_id in self.attachments:
            LOGGER.warning(("Found attachments for post or page #{0}, but didn't find post or page. " +
                            "(Attachments: {1})").format(post_id, [e['files'][0] for e in self.attachments[post_id].values()]))


def get_text_tag(tag, name, default):
    """Get the text of an XML tag."""
    if tag is None:
        return default
    t = tag.find(name)
    if t is not None and t.text is not None:
        return t.text
    else:
        return default


def separate_qtranslate_tagged_langs(text):
    """Parse the content of a wordpress post or page and separate languages.

    For qtranslateX tags: [:LL]blabla[:]

    Note: qtranslate* plugins had a troubled history and used various
    tags over time, application of the 'modernize_qtranslate_tags'
    function is required for this function to handle most of the legacy
    cases.
    """
    qt_start = "[:"
    qt_end = "]"
    qt_end_len = len(qt_end)
    qt_end_with_lang_len = qt_end_len + 2
    qt_chunks = text.split(qt_start)
    content_by_lang = {}
    common_txt_list = []
    for c in qt_chunks:
        if not c.strip():
            continue
        if c.startswith(qt_end):
            # just after the end of a language specific section, there may
            # be some piece of common text or tags, or just nothing
            lang = ""  # default language
            c = c.lstrip(qt_end)
            if not c.strip():
                continue
        elif c[2:qt_end_with_lang_len].startswith(qt_end):
            # a language specific section (with language code at the beginning)
            lang = c[:2]
            c = c[qt_end_with_lang_len:]
        else:
            # nowhere specific (maybe there is no language section in the
            # currently parsed content)
            lang = ""  # default language
        if not lang:
            common_txt_list.append(c)
            for l in content_by_lang.keys():
                content_by_lang[l].append(c)
        else:
            content_by_lang[lang] = content_by_lang.get(lang, common_txt_list) + [c]
    # in case there was no language specific section, just add the text
    if common_txt_list and not content_by_lang:
        content_by_lang[""] = common_txt_list
    # Format back the list to simple text
    for l in content_by_lang.keys():
        content_by_lang[l] = " ".join(content_by_lang[l])
    return content_by_lang


def modernize_qtranslate_tags(xml_bytes):
    """
    Uniformize the "tag" used by various version of qtranslate.

    The resulting byte string will only contain one set of qtranslate tags
    (namely [:LG] and [:]), older ones being converted to new ones.
    """
    old_start_lang = re.compile(b"<!--:?(\\w{2})-->")
    new_start_lang = b"[:\\1]"
    old_end_lang = re.compile(b"<!--(/\\w{2}|:)-->")
    new_end_lang = b"[:]"
    title_match = re.compile(b"<title>(.*?)</title>")
    modern_starts = old_start_lang.sub(new_start_lang, xml_bytes)
    modernized_bytes = old_end_lang.sub(new_end_lang, modern_starts)

    def title_escape(match):
        title = match.group(1)
        title = title.replace(b"&", b"&amp;").replace(b"<", b"&lt;").replace(b">", b"&gt;")
        return b"<title>" + title + b"</title>"
    fixed_bytes = title_match.sub(title_escape, modernized_bytes)
    return fixed_bytes



# -*- coding: utf-8 -*-

# Copyright © 2012-2024 Roberto Alsina and others.

# Permission is hereby granted, free of charge, to any
# person obtaining a copy of this software and associated
# documentation files (the "Software"), to deal in the
# Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the
# Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice
# shall be included in all copies or substantial portions of
# the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY
# KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
# PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS
# OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR
# OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
# OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

"""Utility functions."""

import configparser
import datetime
import hashlib
import io
import lxml.html
import operator
import os
import re
import json
import shutil
import socket
import subprocess
import sys
import threading
import typing
from collections import defaultdict, OrderedDict
from collections.abc import Callable, Iterable
from html import unescape as html_unescape
from importlib import reload as _reload
from unicodedata import normalize as unicodenormalize
from urllib.parse import quote as urlquote
from urllib.parse import unquote as urlunquote
from urllib.parse import urlparse, urlunparse
from zipfile import ZipFile

import babel.dates
import dateutil.parser
import dateutil.tz
import pygments.formatters
import pygments.formatters._mapping
import PyRSS2Gen as rss
from blinker import signal
from doit import tools
from doit.cmdparse import CmdParse
from nikola.packages.pygments_better_html import BetterHtmlFormatter
from typing import List
from unidecode import unidecode

# Renames
from nikola import DEBUG  # NOQA
from .log import LOGGER, get_logger  # NOQA
from .hierarchy_utils import TreeNode, clone_treenode, flatten_tree_structure, sort_classifications
from .hierarchy_utils import join_hierarchical_category_path, parse_escaped_hierarchical_category_name

if sys.version_info.minor <= 8:
    from pkg_resources import resource_filename
else:
    from importlib import resources

try:
    import toml
except ImportError:
    toml = None

try:
    from ruamel.yaml import YAML
except ImportError:
    YAML = None

try:
    import hsluv
except ImportError:
    hsluv = None

__all__ = ('CustomEncoder', 'get_theme_path', 'get_theme_path_real',
           'get_theme_chain', 'load_messages', 'copy_tree', 'copy_file',
           'slugify', 'unslugify', 'to_datetime', 'apply_filters',
           'config_changed', 'get_crumbs', 'get_tzname', 'get_asset_path',
           '_reload', 'Functionary', 'TranslatableSetting',
           'TemplateHookRegistry', 'LocaleBorg',
           'sys_encode', 'sys_decode', 'makedirs', 'get_parent_theme_name',
           'demote_headers', 'get_translation_candidate', 'write_metadata',
           'ask', 'ask_yesno', 'options2docstring', 'os_path_split',
           'get_displayed_page_number', 'adjust_name_for_index_path_list',
           'adjust_name_for_index_path', 'adjust_name_for_index_link',
           'NikolaPygmentsHTML', 'create_redirect', 'clean_before_deployment',
           'sort_posts', 'smartjoin', 'indent', 'load_data', 'html_unescape',
           'rss_writer', 'map_metadata', 'req_missing', 'bool_from_meta',
           # Deprecated, moved to hierarchy_utils:
           'TreeNode', 'clone_treenode', 'flatten_tree_structure',
           'sort_classifications', 'join_hierarchical_category_path',
           'parse_escaped_hierarchical_category_name',)

# Are you looking for 'generic_rss_renderer'?
# It's defined in nikola.nikola.Nikola (the site object).

# Aliases, previously for Python 2/3 compatibility.
# TODO remove in v9
bytes_str = bytes
unicode_str = str
unichr = chr

# For compatibility with old logging setups.
# TODO remove in v9?
STDERR_HANDLER = None


USE_SLUGIFY = True


def req_missing(names, purpose, python=True, optional=False):
    """Log that we are missing some requirements.

    `names` is a list/tuple/set of missing things.
    `purpose` is a string, specifying the use of the missing things.
              It completes the sentence:
                  In order to {purpose}, you must install ...
    `python` specifies whether the requirements are Python packages
                               or other software.
    `optional` specifies whether the things are required
                                 (this is an error and we exit with code 5)
                                 or not (this is just a warning).

    Returns the message shown to the user (which you can usually discard).
    If no names are specified, False is returned and nothing is shown
    to the user.

    """
    if not (isinstance(names, tuple) or isinstance(names, list) or isinstance(names, set)):
        names = (names,)
    if not names:
        return False
    if python:
        whatarethey_s = 'Python package'
        whatarethey_p = 'Python packages'
    else:
        whatarethey_s = whatarethey_p = 'software'
    if len(names) == 1:
        msg = 'In order to {0}, you must install the "{1}" {2}.'.format(
            purpose, names[0], whatarethey_s)
    else:
        most = '", "'.join(names[:-1])
        pnames = most + '" and "' + names[-1]
        msg = 'In order to {0}, you must install the "{1}" {2}.'.format(
            purpose, pnames, whatarethey_p)

    if optional:
        LOGGER.warning(msg)
    else:
        LOGGER.error(msg)
        LOGGER.error('Exiting due to missing dependencies.')
        sys.exit(5)

    return msg


ENCODING = sys.getfilesystemencoding() or sys.stdin.encoding


def sys_encode(thing):
    """Return bytes encoded in the system's encoding."""
    if isinstance(thing, str):
        return thing.encode(ENCODING)
    return thing


def sys_decode(thing):
    """Return Unicode."""
    if isinstance(thing, bytes):
        return thing.decode(ENCODING)
    return thing


def makedirs(path):
    """Create a folder and its parents if needed (mkdir -p)."""
    if not path:
        return
    if os.path.exists(path):
        if not os.path.isdir(path):
            raise OSError('Path {0} already exists and is not a folder.'.format(path))
        else:
            return
    try:
        os.makedirs(path)
        return
    except Exception:
        if os.path.isdir(path):
            return
        raise


class Functionary(defaultdict):
    """Class that looks like a function, but is a defaultdict."""

    def __init__(self, default, default_lang):
        """Initialize a functionary."""
        super().__init__(default)
        self.default_lang = default_lang

    def __call__(self, key, lang=None):
        """When called as a function, take an optional lang and return self[lang][key]."""
        if lang is None:
            lang = LocaleBorg().current_lang
        return self[lang][key]


class TranslatableSetting(object):
    """A setting that can be translated.

    You can access it via: SETTING(lang).  You can omit lang, in which
    case Nikola will ask LocaleBorg, unless you set SETTING.lang,
    which overrides that call.

    You can also stringify the setting and you will get something
    sensible (in what LocaleBorg claims the language is, can also be
    overriden by SETTING.lang). Note that this second method is
    deprecated.  It is kept for backwards compatibility and
    safety.  It is not guaranteed.

    The underlying structure is a defaultdict.  The language that
    is the default value of the dict is provided with __init__().
    """

    # WARNING: This is generally not used and replaced with a call to
    #          LocaleBorg().  Set this to a truthy value to override that.
    lang = None

    # Note that this setting is global.  DO NOT set on a per-instance basis!
    default_lang = 'en'

    def __getattribute__(self, attr):
        """Return attributes, falling back to string attributes."""
        try:
            return super().__getattribute__(attr)
        except AttributeError:
            return self().__getattribute__(attr)

    def __dir__(self):
        """Return the available methods of TranslatableSettings and strings."""
        return list(set(self.__dict__).union(set(dir(str))))

    def __init__(self, name, inp, translations):
        """Initialize a translated setting.

        Valid inputs include:

        * a string               -- the same will be used for all languages
        * a dict ({lang: value}) -- each language will use the value specified;
                                    if there is none, default_lang is used.
        """
        self.name = name
        self._inp = inp
        self.translations = translations
        self.overriden_default = False
        self.values = defaultdict()

        if isinstance(inp, dict) and inp:
            self.translated = True
            self.values.update(inp)
            if self.default_lang not in self.values.keys():
                self.default_lang = list(self.values.keys())[0]
                self.overridden_default = True
            self.values.default_factory = lambda: self.values[self.default_lang]
            for k in translations.keys():
                if k not in self.values.keys():
                    self.values[k] = inp[self.default_lang]
        else:
            self.translated = False
            self.values[self.default_lang] = inp
            self.values.default_factory = lambda: inp

    def get_lang(self):
        """Return the language that should be used to retrieve settings."""
        if self.lang:
            return self.lang
        elif not self.translated:
            return self.default_lang
        else:
            try:
                return LocaleBorg().current_lang
            except AttributeError:
                return self.default_lang

    def __call__(self, lang=None):
        """Return the value in the requested language.

        While lang is None, self.lang (currently set language) is used.
        Otherwise, the standard algorithm is used (see above).

        """
        if lang is None:
            return self.values[self.get_lang()]
        else:
            return self.values[lang]

    def __str__(self):
        """Return the value in the currently set language (deprecated)."""
        return str(self.values[self.get_lang()])

    def __repr__(self):
        """Provide a representation for programmers."""
        return '<TranslatableSetting: {0!r} = {1!r}>'.format(self.name, self._inp)

    def format(self, *args, **kwargs):
        """Format ALL the values in the setting the same way."""
        for l in self.values:
            self.values[l] = self.values[l].format(*args, **kwargs)
        self.values.default_factory = lambda: self.values[self.default_lang]
        return self

    def langformat(self, formats):
        """Format ALL the values in the setting, on a per-language basis."""
        if not formats:
            # Input is empty.
            return self
        else:
            # This is a little tricky.
            # Basically, we have some things that may very well be dicts.  Or
            # actually, TranslatableSettings in the original unprocessed dict
            # form.  We need to detect them.

            # First off, we need to check what languages we have and what
            # should we use as the default.
            keys = list(formats)
            if self.default_lang in keys:
                d = formats[self.default_lang]
            else:
                d = formats[keys[0]]
            # Discovering languages of the settings here.
            langkeys = []
            for f in formats.values():
                for a in f[0] + tuple(f[1].values()):
                    if isinstance(a, dict):
                        langkeys += list(a)

            # Now that we know all this, we go through all the languages we have.
            allvalues = set(keys + langkeys + list(self.values))
            self.values['__orig__'] = self.values[self.default_lang]
            for l in allvalues:
                if l in keys:
                    oargs, okwargs = formats[l]
                else:
                    oargs, okwargs = d

                args = []
                kwargs = {}

                for a in oargs:
                    # We create temporary TranslatableSettings and replace the
                    # values with them.
                    if isinstance(a, dict):
                        a = TranslatableSetting('NULL', a, self.translations)
                        args.append(a(l))
                    else:
                        args.append(a)

                for k, v in okwargs.items():
                    if isinstance(v, dict):
                        v = TranslatableSetting('NULL', v, self.translations)
                        kwargs.update({k: v(l)})
                    else:
                        kwargs.update({k: v})

                if l in self.values:
                    self.values[l] = self.values[l].format(*args, **kwargs)
                else:
                    self.values[l] = self.values['__orig__'].format(*args, **kwargs)
                self.values.default_factory = lambda: self.values[self.default_lang]

        return self

    def __getitem__(self, key):
        """Provide an alternate interface via __getitem__."""
        return self.values[key]

    def __setitem__(self, key, value):
        """Set values for translations."""
        self.values[key] = value

    def __eq__(self, other):
        """Test whether two TranslatableSettings are equal."""
        try:
            return self.values == other.values
        except AttributeError:
            return self(self.default_lang) == other

    def __ne__(self, other):
        """Test whether two TranslatableSettings are inequal."""
        try:
            return self.values != other.values
        except AttributeError:
            return self(self.default_lang) != other


class TemplateHookRegistry(object):
    r"""A registry for template hooks.

    Usage:

    >>> r = TemplateHookRegistry('foo', None)
    >>> r.append('Hello!')
    >>> r.append(lambda x: 'Hello ' + x + '!', False, 'world')
    >>> repr(r())
    'Hello!\nHello world!'
    """

    def __init__(self, name, site):
        """Initialize a hook registry."""
        self._items = []
        self.name = name
        self.site = site
        self.context = None

    def generate(self):
        """Generate items."""
        for c, inp, site, args, kwargs in self._items:
            if c:
                if site:
                    kwargs['site'] = self.site
                    kwargs['context'] = self.context
                yield inp(*args, **kwargs)
            else:
                yield inp

    def __call__(self):
        """Return items, in a string, separated by newlines."""
        return '\n'.join(self.generate())

    def append(self, inp, wants_site_and_context=False, *args, **kwargs):
        """
        Register an item.

        `inp` can be a string or a callable returning one.
        `wants_site` tells whether there should be a `site` keyword
                     argument provided, for accessing the site.

        Further positional and keyword arguments are passed as-is to the
        callable.

        `wants_site`, args and kwargs are ignored (but saved!) if `inp`
        is not callable.  Callability of `inp` is determined only once.
        """
        c = callable(inp)
        self._items.append((c, inp, wants_site_and_context, args, kwargs))

    def calculate_deps(self):
        """Calculate dependencies for a registry."""
        deps = []
        for is_callable, inp, wants_site_and_context, args, kwargs in self._items:
            if not is_callable:
                name = inp
            elif hasattr(inp, 'template_registry_identifier'):
                name = inp.template_registry_identifier
            elif hasattr(inp, '__doc__'):
                name = inp.__doc__
            else:
                name = '_undefined_callable_'
            deps.append((is_callable, name, wants_site_and_context, args, kwargs))

    def __hash__(self):
        """Return hash of a registry."""
        return hash(config_changed({self.name: self.calculate_deps()})._calc_digest())

    def __str__(self):
        """Stringify a registry."""
        return '<TemplateHookRegistry: {0}>'.format(self._items)

    def __repr__(self):
        """Provide the representation of a registry."""
        return '<TemplateHookRegistry: {0}>'.format(self.name)


class CustomEncoder(json.JSONEncoder):
    """Custom JSON encoder."""

    def default(self, obj):
        """Create default encoding handler."""
        try:
            return super().default(obj)
        except TypeError:
            if isinstance(obj, (set, frozenset)):
                return self.encode(sorted(list(obj)))
            elif isinstance(obj, TranslatableSetting):
                s = json.dumps(obj._inp, cls=CustomEncoder, sort_keys=True)
            else:
                s = repr(obj).split('0x', 1)[0]
            return s


class config_changed(tools.config_changed):
    """A copy of doit's config_changed, using pickle instead of serializing manually."""

    def __init__(self, config, identifier=None):
        """Initialize config_changed."""
        super().__init__(config)
        self.identifier = '_config_changed'
        if identifier is not None:
            self.identifier += ':' + identifier

    # DEBUG (for unexpected rebuilds)
    @classmethod
    def _write_into_debug_db(cls, digest: str, data: str) -> None:  # pragma: no cover
        """Write full values of config_changed into a sqlite3 database."""
        import sqlite3
        try:
            cls.debug_db_cursor
        except AttributeError:
            cls.debug_db_conn = sqlite3.connect("cc_debug.sqlite3")
            cls.debug_db_id = datetime.datetime.now().isoformat()
            cls.debug_db_cursor = cls.debug_db_conn.cursor()
            cls.debug_db_cursor.execute("""
            CREATE TABLE IF NOT EXISTS hashes (hash CHARACTER(32) PRIMARY KEY, json_data TEXT);
            """)
            cls.debug_db_conn.commit()

        try:
            cls.debug_db_cursor.execute("INSERT INTO hashes (hash, json_data) VALUES (?, ?);", (digest, data))
            cls.debug_db_conn.commit()
        except sqlite3.IntegrityError:
            # ON CONFLICT DO NOTHING, except Ubuntu 16.04’s sqlite3 is too ancient for this
            cls.debug_db_conn.rollback()

    def _calc_digest(self):
        """Calculate a config_changed digest."""
        if isinstance(self.config, str):
            return self.config
        elif isinstance(self.config, dict):
            data = json.dumps(self.config, cls=CustomEncoder, sort_keys=True)
            if isinstance(data, str):  # pragma: no cover # python3
                byte_data = data.encode("utf-8")
            else:
                byte_data = data
            digest = hashlib.md5(byte_data).hexdigest()

            # DEBUG (for unexpected rebuilds)
            # self._write_into_debug_db(digest, data)
            # Alternative (without database):
            # LOGGER.debug('{{"{0}": {1}}}'.format(digest, byte_data))
            # Humanized format:
            # LOGGER.debug('[Digest {0} for {2}]\n{1}\n[Digest {0} for {2}]'.format(digest, byte_data, self.identifier))

            return digest
        else:
            raise Exception('Invalid type of config_changed parameter -- got '
                            '{0}, must be string or dict'.format(type(
                                self.config)))

    def configure_task(self, task):
        """Configure a task with a digest."""
        task.value_savers.append(lambda: {self.identifier: self._calc_digest()})

    def __call__(self, task, values):
        """Return True if config values are unchanged."""
        last_success = values.get(self.identifier)
        if last_success is None:
            return False
        return (last_success == self._calc_digest())

    def __repr__(self):
        """Provide a representation of config_changed."""
        return "Change with config: {0}".format(json.dumps(self.config,
                                                           cls=CustomEncoder,
                                                           sort_keys=True))


def pkg_resources_path(package, resource):
    """Return path to a resource from the package with the given name."""
    if sys.version_info.minor <= 8:
        return resource_filename(package, resource)
    else:
        return str(resources.files(package).joinpath(resource))


def get_theme_path_real(theme, themes_dirs):
    """Return the path where the given theme's files are located.

    Looks in ./themes and in the place where themes go when installed.
    """
    for themes_dir in themes_dirs:
        dir_name = os.path.join(themes_dir, theme)
        if os.path.isdir(dir_name):
            return dir_name
    dir_name = pkg_resources_path('nikola', os.path.join('data', 'themes', theme))
    if os.path.isdir(dir_name):
        return dir_name
    raise Exception("Can't find theme '{0}'".format(theme))


def get_theme_path(theme):
    """Return the theme's path, which equals the theme's name."""
    return theme


def parse_theme_meta(theme_dir):
    """Parse a .theme meta file."""
    cp = configparser.ConfigParser()
    # The `or` case is in case theme_dir ends with a trailing slash
    theme_name = os.path.basename(theme_dir) or os.path.basename(os.path.dirname(theme_dir))
    theme_meta_path = os.path.join(theme_dir, theme_name + '.theme')
    cp.read(theme_meta_path)
    return cp if cp.has_section('Theme') else None


def get_template_engine(themes):
    """Get template engine used by a given theme."""
    for theme_name in themes:
        meta = parse_theme_meta(theme_name)
        if meta:
            e = meta.get('Theme', 'engine', fallback=None)
            if e:
                return e
        else:
            # Theme still uses old-style parent/engine files
            engine_path = os.path.join(theme_name, 'engine')
            if os.path.isfile(engine_path):
                with open(engine_path) as fd:
                    return fd.readlines()[0].strip()
    # default
    return 'mako'


def get_parent_theme_name(theme_name, themes_dirs=None):
    """Get name of parent theme."""
    meta = parse_theme_meta(theme_name)
    if meta:
        parent = meta.get('Theme', 'parent', fallback=None)
        if themes_dirs and parent:
            return get_theme_path_real(parent, themes_dirs)
        return parent
    else:
        # Theme still uses old-style parent/engine files
        parent_path = os.path.join(theme_name, 'parent')
        if os.path.isfile(parent_path):
            with open(parent_path) as fd:
                parent = fd.readlines()[0].strip()
            if themes_dirs:
                return get_theme_path_real(parent, themes_dirs)
            return parent
        return None


def get_theme_chain(theme, themes_dirs):
    """Create the full theme inheritance chain including paths."""
    themes = [get_theme_path_real(theme, themes_dirs)]

    while True:
        parent = get_parent_theme_name(themes[-1], themes_dirs=themes_dirs)
        # Avoid silly loops
        if parent is None or parent in themes:
            break
        themes.append(parent)
    return themes


def html_tostring_fragment(document):
    """Convert a HTML snippet to a fragment, ready for insertion elsewhere."""
    try:
        doc = lxml.html.tostring(document.body, encoding='unicode').strip()
    except Exception:
        doc = lxml.html.tostring(document, encoding='unicode').strip()
    start_fragments = ["<html>", "<body>"]
    end_fragments = ["</body>", "</html>"]
    for start in start_fragments:
        if doc.startswith(start):
            doc = doc[len(start):].strip()
    for end in end_fragments:
        if doc.endswith(end):
            doc = doc[:-len(end)].strip()
    return doc


INCOMPLETE_LANGUAGES_WARNED = set()


class LanguageNotFoundError(Exception):
    """An exception thrown if language is not found."""

    def __init__(self, lang, orig):
        """Initialize exception."""
        self.lang = lang
        self.orig = orig

    def __str__(self):
        """Stringify the exception."""
        return 'cannot find language {0}'.format(self.lang)


def load_messages(themes, translations, default_lang, themes_dirs):
    """Load theme's messages into context.

    All the messages from parent themes are loaded,
    and "younger" themes have priority.
    """
    messages = Functionary(dict, default_lang)
    oldpath = list(sys.path)
    found = {lang: False for lang in translations.keys()}
    last_exception = None
    completion_status = {lang: False for lang in translations.keys()}
    for theme_name in themes[::-1]:
        msg_folder = os.path.join(get_theme_path(theme_name), 'messages')
        default_folder = os.path.join(get_theme_path_real('base', themes_dirs), 'messages')
        sys.path.insert(0, default_folder)
        sys.path.insert(0, msg_folder)

        english = __import__('messages_en')
        # If we don't do the reload, the module is cached
        _reload(english)
        for lang in translations.keys():
            try:
                translation = __import__('messages_' + lang)
                # If we don't do the reload, the module is cached
                _reload(translation)
                found[lang] = True
                if sorted(translation.MESSAGES.keys()) != sorted(english.MESSAGES.keys()):
                    completion_status[lang] = completion_status[lang] or False
                else:
                    completion_status[lang] = True

                messages[lang].update(english.MESSAGES)
                for k, v in translation.MESSAGES.items():
                    if v:
                        messages[lang][k] = v
                del translation
            except ImportError as orig:
                last_exception = orig
        del english
        sys.path = oldpath

    if not all(found.values()):
        raise LanguageNotFoundError(lang, last_exception)
    for lang, status in completion_status.items():
        if not status and lang not in INCOMPLETE_LANGUAGES_WARNED:
            LOGGER.warning("Incomplete translation for language '{0}'.".format(lang))
            INCOMPLETE_LANGUAGES_WARNED.add(lang)

    return messages


def copy_tree(src, dst, link_cutoff=None, ignored_filenames=None):
    """Copy a src tree to the dst folder.

    Example:

    src = "themes/default/assets"
    dst = "output/assets"

    should copy "themes/defauts/assets/foo/bar" to
    "output/assets/foo/bar"

    If link_cutoff is set, then the links pointing at things
    *inside* that folder will stay as links, and links
    pointing *outside* that folder will be copied.

    ignored_filenames is a set of file names that will be ignored.
    """
    ignore = set(['.svn', '.git']) | (ignored_filenames or set())
    base_len = len(src.split(os.sep))
    for root, dirs, files in os.walk(src, followlinks=True):
        root_parts = root.split(os.sep)
        if set(root_parts) & ignore:
            continue
        dst_dir = os.path.join(dst, *root_parts[base_len:])
        makedirs(dst_dir)
        for src_name in files:
            if src_name in ('.DS_Store', 'Thumbs.db'):
                continue
            dst_file = os.path.join(dst_dir, src_name)
            src_file = os.path.join(root, src_name)
            yield {
                'name': dst_file,
                'file_dep': [src_file],
                'targets': [dst_file],
                'actions': [(copy_file, (src_file, dst_file, link_cutoff))],
                'clean': True,
            }


def copy_file(source, dest, cutoff=None):
    """Copy a file from source to dest. If link target starts with `cutoff`, symlinks are used."""
    dst_dir = os.path.dirname(dest)
    makedirs(dst_dir)
    if os.path.islink(source):
        link_target = os.path.relpath(
            os.path.normpath(os.path.join(dst_dir, os.readlink(source))))
        # Now we have to decide if we copy the link target or the
        # link itself.
        if cutoff is None or not link_target.startswith(cutoff):
            # We copy
            shutil.copy2(source, dest)
        else:
            # We link
            if os.path.exists(dest) or os.path.islink(dest):
                os.unlink(dest)
            os.symlink(os.readlink(source), dest)
    else:
        shutil.copy2(source, dest)


def remove_file(source):
    """Remove file or directory."""
    if os.path.isdir(source):
        shutil.rmtree(source)
    elif os.path.isfile(source) or os.path.islink(source):
        os.remove(source)


# slugify is adopted from
# https://code.activestate.com/recipes/
# 577257-slugify-make-a-string-usable-in-a-url-or-filename/
_slugify_strip_re = re.compile(r'[^+\w\s-]', re.UNICODE)
_slugify_hyphenate_re = re.compile(r'[-\s]+', re.UNICODE)


def slugify(value, lang=None, force=False):
    u"""Normalize string, convert to lowercase, remove non-alpha characters, convert spaces to hyphens.

    From Django's "django/template/defaultfilters.py".

    >>> print(slugify('áéí.óú', lang='en'))
    aeiou

    >>> print(slugify('foo/bar', lang='en'))
    foobar

    >>> print(slugify('foo bar', lang='en'))
    foo-bar
    """
    if not isinstance(value, str):
        raise ValueError("Not a unicode object: {0}".format(value))
    if USE_SLUGIFY or force:
        # This is the standard state of slugify, which actually does some work.
        # It is the preferred style, especially for Western languages.
        value = str(unidecode(value))
        value = _slugify_strip_re.sub('', value).strip().lower()
        return _slugify_hyphenate_re.sub('-', value)
    else:
        # This is the “disarmed” state of slugify, which lets the user
        # have any character they please (be it regular ASCII with spaces,
        # or another alphabet entirely).  This might be bad in some
        # environments, and as such, USE_SLUGIFY is better off being True!

        # We still replace some characters, though.  In particular, we need
        # to replace ? and #, which should not appear in URLs, and some
        # Windows-unsafe characters.  This list might be even longer.
        rc = '/\\?#"\'\r\n\t*:<>|'

        for c in rc:
            value = value.replace(c, '-')
        return value


def unslugify(value, lang=None, discard_numbers=True):
    """Given a slug string (as a filename), return a human readable string.

    If discard_numbers is True, numbers right at the beginning of input
    will be removed.
    """
    if discard_numbers:
        value = re.sub('^[0-9]+', '', value)
    value = re.sub(r'([_\-\.])', ' ', value)
    value = value.strip().capitalize()
    return value


def encodelink(iri):
    """Given an encoded or unencoded link string, return an encoded string suitable for use as a link in HTML and XML."""
    iri = unicodenormalize('NFC', iri)
    link = OrderedDict(urlparse(iri)._asdict())
    link['path'] = urlquote(urlunquote(link['path']).encode('utf-8'), safe="/~")
    try:
        link['netloc'] = link['netloc'].encode('utf-8').decode('idna').encode('idna').decode('utf-8')
    except UnicodeDecodeError:
        link['netloc'] = link['netloc'].encode('idna').decode('utf-8')
    encoded_link = urlunparse(link.values())
    return encoded_link


def full_path_from_urlparse(parsed) -> str:
    """Given urlparse output, return the full path (with query and fragment)."""
    dst = parsed.path
    if parsed.query:
        dst = "{0}?{1}".format(dst, parsed.query)
    if parsed.fragment:
        dst = "{0}#{1}".format(dst, parsed.fragment)
    return dst

# A very slightly safer version of zip.extractall that works on
# python < 2.6


class UnsafeZipException(Exception):
    """Exception for unsafe zip files."""

    pass


def extract_all(zipfile, path='themes'):
    """Extract all files from a zip file."""
    pwd = os.getcwd()
    makedirs(path)
    try:
        os.chdir(path)
        z = ZipFile(zipfile)
        namelist = z.namelist()
        for f in namelist:
            if f.endswith('/') and '..' in f:
                raise UnsafeZipException('The zip file contains ".." and is '
                                         'not safe to expand.')
        for f in namelist:
            if f.endswith('/'):
                makedirs(f)
            else:
                z.extract(f)
        z.close()
    finally:
        os.chdir(pwd)


def to_datetime(value, tzinfo=None):
    """Convert string to datetime."""
    try:
        if type(value) is datetime.date:
            # type() instead of isinstance() is expected here, since we don’t
            # want to change datetime.datetime objects.
            value = datetime.datetime.combine(value, datetime.time(0, 0))
        if not isinstance(value, datetime.datetime):
            # dateutil does bad things with TZs like UTC-03:00.
            dateregexp = re.compile(r' UTC([+-][0-9][0-9]:[0-9][0-9])')
            value = re.sub(dateregexp, r'\1', value)
            value = dateutil.parser.parse(value)
        if not value.tzinfo:
            value = value.replace(tzinfo=tzinfo)
        return value
    except Exception:
        raise ValueError('Unrecognized date/time: {0!r}'.format(value))


def get_tzname(dt):
    """Given a datetime value, find the name of the time zone.

    DEPRECATED: This thing returned basically the 1st random zone
    that matched the offset.
    """
    return dt.tzname()


def current_time(tzinfo=None):
    """Get current time."""
    if tzinfo is not None:
        dt = datetime.datetime.now(tzinfo)
    else:
        dt = datetime.datetime.now(dateutil.tz.tzlocal())
    return dt


from nikola import filters as task_filters  # NOQA


def apply_filters(task, filters, skip_ext=None):
    """Apply filters to a task.

    If any of the targets of the given task has a filter that matches,
    adds the filter commands to the commands of the task,
    and the filter itself to the uptodate of the task.
    """
    if '.php' in filters.keys():
        if task_filters.php_template_injection not in filters['.php']:
            filters['.php'].append(task_filters.php_template_injection)
    else:
        filters['.php'] = [task_filters.php_template_injection]

    def filter_matches(ext):
        for key, value in list(filters.items()):
            if isinstance(key, (tuple, list)):
                if ext in key:
                    return value
            elif isinstance(key, (bytes, str)):
                if ext == key:
                    return value
            else:
                raise ValueError("Cannot find filter match for {0}".format(key))

    for target in task.get('targets', []):
        ext = os.path.splitext(target)[-1].lower()
        if skip_ext and ext in skip_ext:
            continue
        filter_ = filter_matches(ext)
        if filter_:
            for action in filter_:
                def unlessLink(action, target):
                    if not os.path.islink(target):
                        if isinstance(action, Callable):
                            action(target)
                        else:
                            subprocess.check_call(action % target, shell=True)

                task['actions'].append((unlessLink, (action, target)))
    return task


def get_crumbs(path, is_file=False, index_folder=None, lang=None):
    """Create proper links for a crumb bar.

    index_folder is used if you want to use title from index file
    instead of folder name as breadcrumb text.

    >>> crumbs = get_crumbs('galleries')
    >>> len(crumbs)
    1
    >>> crumbs[0]
    ['#', 'galleries']

    >>> crumbs = get_crumbs(os.path.join('galleries','demo'))
    >>> len(crumbs)
    2
    >>> crumbs[0]
    ['..', 'galleries']
    >>> crumbs[1]
    ['#', 'demo']

    >>> crumbs = get_crumbs(os.path.join('listings','foo','bar'), is_file=True)
    >>> len(crumbs)
    3
    >>> crumbs[0]
    ['..', 'listings']
    >>> crumbs[1]
    ['.', 'foo']
    >>> crumbs[2]
    ['#', 'bar']
    """
    crumbs = path.split(os.sep)
    _crumbs = []
    if is_file:
        for i, crumb in enumerate(crumbs[-3::-1]):  # Up to parent folder only
            _path = '/'.join(['..'] * (i + 1))
            _crumbs.append([_path, crumb])
        if len(crumbs) >= 2:
            _crumbs.insert(0, ['.', crumbs[-2]])  # file's folder
        if len(crumbs) >= 1:
            _crumbs.insert(0, ['#', crumbs[-1]])  # file itself
    else:
        for i, crumb in enumerate(crumbs[::-1]):
            _path = '/'.join(['..'] * i) or '#'
            _crumbs.append([_path, crumb])
    if index_folder and hasattr(index_folder, 'parse_index'):
        folder = path
        for i, crumb in enumerate(crumbs[::-1]):
            if folder[-1] == os.sep:
                folder = folder[:-1]
            # We don't care about the created Post() object except for its title;
            # hence, the input_folder and output_folder given to
            # index_folder.parse_index() don't matter
            index_post = index_folder.parse_index(folder, '', '')
            folder = folder.replace(crumb, '')
            if index_post:
                crumb = index_post.title(lang) or crumb
            _crumbs[i][1] = crumb
    return list(reversed(_crumbs))


def get_asset_path(path, themes, files_folders={'files': ''}, output_dir='output'):
    """Return the "real", absolute path to the asset.

    By default, it checks which theme provides the asset.
    If the asset is not provided by a theme, then it will be checked for
    in the FILES_FOLDERS.
    If it's not provided by either, it will be chacked in output, where
    it may have been created by another plugin.

    >>> print(get_asset_path('assets/css/nikola_rst.css', get_theme_chain('bootstrap3', ['themes'])))
    /.../nikola/data/themes/base/assets/css/nikola_rst.css

    >>> print(get_asset_path('assets/css/theme.css', get_theme_chain('bootstrap3', ['themes'])))
    /.../nikola/data/themes/bootstrap3/assets/css/theme.css

    >>> print(get_asset_path('nikola.py', get_theme_chain('bootstrap3', ['themes']), {'nikola': ''}))
    /.../nikola/nikola.py

    >>> print(get_asset_path('nikola.py', get_theme_chain('bootstrap3', ['themes']), {'nikola': 'nikola'}))
    None

    >>> print(get_asset_path('nikola/nikola.py', get_theme_chain('bootstrap3', ['themes']), {'nikola': 'nikola'}))
    /.../nikola/nikola.py

    """
    for theme_name in themes:
        candidate = os.path.join(get_theme_path(theme_name), path)
        if os.path.isfile(candidate):
            return candidate
    for src, rel_dst in files_folders.items():
        relpath = os.path.normpath(os.path.relpath(path, rel_dst))
        if not relpath.startswith('..' + os.path.sep):
            candidate = os.path.abspath(os.path.join(src, relpath))
            if os.path.isfile(candidate):
                return candidate

    if output_dir:
        candidate = os.path.join(output_dir, path)
        if os.path.isfile(candidate):
            return candidate

    # whatever!
    return None


class LocaleBorgUninitializedException(Exception):
    """Exception for unitialized LocaleBorg."""

    def __init__(self):
        """Initialize exception."""
        super().__init__("Attempt to use LocaleBorg before initialization")


# Customized versions of babel.dates functions that don't do weird stuff with
# timezones. Without these fixes, DST would follow local settings (because
# dateutil’s timezones return stuff depending on their input, and datetime.time
# objects have no year/month/day to base the information on.
def format_datetime(datetime=None, format='medium',
                    locale=babel.dates.LC_TIME):
    """Format a datetime object."""
    locale = babel.dates.Locale.parse(locale)
    if format in ('full', 'long', 'medium', 'short'):
        return babel.dates.get_datetime_format(format, locale=locale) \
            .replace("'", "") \
            .replace('{0}', format_time(datetime, format, locale=locale)) \
            .replace('{1}', babel.dates.format_date(datetime, format, locale=locale))
    else:
        return babel.dates.parse_pattern(format).apply(datetime, locale)


def format_time(time=None, format='medium', locale=babel.dates.LC_TIME):
    """Format time. Input can be datetime.time or datetime.datetime."""
    locale = babel.dates.Locale.parse(locale)
    if format in ('full', 'long', 'medium', 'short'):
        format = babel.dates.get_time_format(format, locale=locale)
    return babel.dates.parse_pattern(format).apply(time, locale)


def format_skeleton(skeleton, datetime=None, fo=None, fuzzy=True,
                    locale=babel.dates.LC_TIME):
    """Format a datetime based on a skeleton."""
    locale = babel.dates.Locale.parse(locale)
    if fuzzy and skeleton not in locale.datetime_skeletons:
        skeleton = babel.dates.match_skeleton(skeleton, locale.datetime_skeletons)
    format = locale.datetime_skeletons[skeleton]
    return format_datetime(datetime, format, locale)


class LocaleBorg(object):
    """Provide locale related services and autoritative current_lang.

    This class stores information about the locales used and interfaces
    with the Babel library to provide internationalization services.

    Usage:
        # early in cmd or test execution
        LocaleBorg.initialize(...)

        # any time later
        lang = LocaleBorg().<service>

    Available services:
        .current_lang: autoritative current_lang, the last seen in set_locale
        .formatted_date: format a date(time) according to locale rules
        .format_date_in_string: take a message and format the date in it

    The default implementation uses the Babel package and completely ignores
    the Python `locale` module. If you wish to override this, write functions
    and assign them to the appropriate names. The functions are:

     * LocaleBorg.datetime_formatter(date, date_format, lang, locale)
     * LocaleBorg.in_string_formatter(date, mode, custom_format, lang, locale)
    """

    initialized = False

    # Can be used to override Babel
    datetime_formatter = None
    in_string_formatter = None

    @classmethod
    def initialize(cls, locales: 'typing.Dict[str, str]', initial_lang: str):
        """Initialize LocaleBorg.

        locales: dict with custom locale name overrides.
        """
        if not initial_lang:
            raise ValueError("Unknown initial language {0}".format(initial_lang))
        cls.reset()
        cls.locales = locales
        cls.__initial_lang = initial_lang
        cls.initialized = True

    def __get_shared_state(self):
        if not self.initialized:  # pragma: no cover
            raise LocaleBorgUninitializedException()
        shared_state = getattr(self.__thread_local, 'shared_state', None)
        if shared_state is None:
            shared_state = {'current_lang': self.__initial_lang}
            self.__thread_local.shared_state = shared_state
        return shared_state

    @classmethod
    def reset(cls):
        """Reset LocaleBorg.

        Used in testing to prevent leaking state between tests.
        """
        cls.__thread_local = threading.local()
        cls.__thread_lock = threading.Lock()

        cls.locales = {}
        cls.initialized = False
        cls.thread_local = None
        cls.datetime_formatter = None
        cls.in_string_formatter = None

    def __init__(self):
        """Initialize."""
        if not self.initialized:
            raise LocaleBorgUninitializedException()

    @property
    def current_lang(self) -> str:
        """Return the current language."""
        return self.__get_shared_state()['current_lang']

    def set_locale(self, lang: str) -> str:
        """Set the current language and return an empty string (to make use in templates easier)."""
        with self.__thread_lock:
            self.__get_shared_state()['current_lang'] = lang
            return ''

    def formatted_date(self, date_format: 'str',
                       date: 'typing.Union[datetime.date, datetime.datetime]',
                       lang: 'typing.Optional[str]' = None) -> str:
        """Return the formatted date/datetime as a string."""
        if lang is None:
            lang = self.current_lang
        locale = self.locales.get(lang, lang)
        # Get a string out of a TranslatableSetting
        if isinstance(date_format, TranslatableSetting):
            date_format = date_format(lang)

        # Always ask Python if the date_format is webiso
        if date_format == 'webiso':
            # Formatted after RFC 3339 (web ISO 8501 profile) with Zulu
            # zone designator for times in UTC and no microsecond precision.
            return date.replace(microsecond=0).isoformat().replace('+00:00', 'Z')
        elif LocaleBorg.datetime_formatter is not None:
            return LocaleBorg.datetime_formatter(date, date_format, lang, locale)
        else:
            return format_datetime(date, date_format, locale=locale)

    def format_date_in_string(self, message: str, date: datetime.date, lang: 'typing.Optional[str]' = None) -> str:
        """Format date inside a string (message).

        Accepted modes: month, month_year, month_day_year.
        Format: {month} for standard, {month:MMMM} for customization.
        """
        modes = {
            'month': ('date', 'LLLL'),
            'month_year': ('skeleton', 'yMMMM'),
            'month_day_year': ('date', 'long')
        }

        if lang is None:
            lang = self.current_lang
        locale = self.locales.get(lang, lang)

        def date_formatter(match: typing.Match) -> str:
            """Format a date as requested."""
            mode, custom_format = match.groups()
            if LocaleBorg.in_string_formatter is not None:
                return LocaleBorg.in_string_formatter(date, mode, custom_format, lang, locale)
            elif custom_format:
                return babel.dates.format_date(date, custom_format, locale)
            else:
                function, fmt = modes[mode]
                if function == 'skeleton':
                    return format_skeleton(fmt, date, locale=locale)
                else:
                    return babel.dates.format_date(date, fmt, locale)

        return re.sub(r'{(.*?)(?::(.*?))?}', date_formatter, message)


class ExtendedRSS2(rss.RSS2):
    """Extended RSS class."""

    xsl_stylesheet_href = None

    def publish(self, handler):
        """Publish a feed."""
        if self.xsl_stylesheet_href:
            handler.processingInstruction("xml-stylesheet", 'type="text/xsl" href="{0}" media="all"'.format(self.xsl_stylesheet_href))
        super().publish(handler)

    def publish_extensions(self, handler):
        """Publish extensions."""
        if self.self_url:
            handler.startElement("atom:link", {
                'href': self.self_url,
                'rel': "self",
                'type': "application/rss+xml"
            })
            handler.endElement("atom:link")


class ExtendedItem(rss.RSSItem):
    """Extended RSS item."""

    def __init__(self, **kw):
        """Initialize RSS item."""
        self.creator = kw.pop('creator', None)

        # It's an old style class
        rss.RSSItem.__init__(self, **kw)

    def publish_extensions(self, handler):
        """Publish extensions."""
        if self.creator:
            handler.startElement("dc:creator", {})
            handler.characters(self.creator)
            handler.endElement("dc:creator")


# \x00 means the "<" was backslash-escaped
explicit_title_re = re.compile(r'^(.+?)\s*(?<!\x00)<(.*?)>$', re.DOTALL)


def split_explicit_title(text):
    """Split role content into title and target, if given.

    From Sphinx's "sphinx/util/nodes.py"
    """
    match = explicit_title_re.match(text)
    if match:
        return True, match.group(1), match.group(2)
    return False, text, text


def first_line(doc):
    """Extract first non-blank line from text, to extract docstring title."""
    if doc is not None:
        for line in doc.splitlines():
            striped = line.strip()
            if striped:
                return striped
    return ''


def demote_headers(doc, level=1):
    """Demote <hN> elements by one."""
    if level == 0:
        return doc
    elif level > 0:
        levels = range(1, 7 - (level - 1))
        levels = reversed(levels)
    elif level < 0:
        levels = range(2 + level, 7)

    for before in levels:
        after = before + level
        if after < 1:
            # html headers can't go lower than 1
            after = 1
        elif after > 6:
            # html headers go until 6
            after = 6

        if before == after:
            continue

        elements = doc.xpath('//h{}'.format(before))
        new_tag = 'h{}'.format(after)
        for element in elements:
            element.tag = new_tag


def get_root_dir():
    """Find root directory of nikola site by looking for conf.py."""
    root = os.getcwd()

    confname = 'conf.py'

    while True:
        if os.path.exists(os.path.join(root, confname)):
            return root
        else:
            basedir = os.path.split(root)[0]
            # Top directory, already checked
            if basedir == root:
                break
            root = basedir

    return None


def get_translation_candidate(config, path, lang):
    """Return a possible path where we can find the translated version of some page, based on the TRANSLATIONS_PATTERN configuration variable.

    >>> config = {'TRANSLATIONS_PATTERN': '{path}.{lang}.{ext}', 'DEFAULT_LANG': 'en', 'TRANSLATIONS': {'es':'1', 'en': 1}}
    >>> print(get_translation_candidate(config, '*.rst', 'es'))
    *.es.rst
    >>> print(get_translation_candidate(config, 'fancy.post.rst', 'es'))
    fancy.post.es.rst
    >>> print(get_translation_candidate(config, '*.es.rst', 'es'))
    *.es.rst
    >>> print(get_translation_candidate(config, '*.es.rst', 'en'))
    *.rst
    >>> print(get_translation_candidate(config, 'cache/posts/fancy.post.es.html', 'en'))
    cache/posts/fancy.post.html
    >>> print(get_translation_candidate(config, 'cache/posts/fancy.post.html', 'es'))
    cache/posts/fancy.post.es.html
    >>> print(get_translation_candidate(config, 'cache/pages/charts.html', 'es'))
    cache/pages/charts.es.html
    >>> print(get_translation_candidate(config, 'cache/pages/charts.html', 'en'))
    cache/pages/charts.html

    >>> config = {'TRANSLATIONS_PATTERN': '{path}.{ext}.{lang}', 'DEFAULT_LANG': 'en', 'TRANSLATIONS': {'es':'1', 'en': 1}}
    >>> print(get_translation_candidate(config, '*.rst', 'es'))
    *.rst.es
    >>> print(get_translation_candidate(config, '*.rst.es', 'es'))
    *.rst.es
    >>> print(get_translation_candidate(config, '*.rst.es', 'en'))
    *.rst
    >>> print(get_translation_candidate(config, 'cache/posts/fancy.post.html.es', 'en'))
    cache/posts/fancy.post.html
    >>> print(get_translation_candidate(config, 'cache/posts/fancy.post.html', 'es'))
    cache/posts/fancy.post.html.es
    """
    # FIXME: this is rather slow and this function is called A LOT
    # Convert the pattern into a regexp
    pattern = config['TRANSLATIONS_PATTERN']
    # This will still break if the user has ?*[]\ in the pattern. But WHY WOULD HE?
    pattern = pattern.replace('.', r'\.')
    pattern = pattern.replace('{path}', '(?P<path>.+?)')
    pattern = pattern.replace('{ext}', r'(?P<ext>[^\./]+)')
    pattern = pattern.replace('{lang}', '(?P<lang>{0})'.format('|'.join(config['TRANSLATIONS'].keys())))
    m = re.match(pattern, path)
    if m and all(m.groups()):  # It's a translated path
        p, e, l = m.group('path'), m.group('ext'), m.group('lang')
        if l == lang:  # Nothing to do
            return path
        elif lang == config['DEFAULT_LANG']:  # Return untranslated path
            return '{0}.{1}'.format(p, e)
        else:  # Change lang and return
            return config['TRANSLATIONS_PATTERN'].format(path=p, ext=e, lang=lang)
    else:
        # It's a untranslated path, assume it's path.ext
        p, e = os.path.splitext(path)
        e = e[1:]  # No initial dot
        if lang == config['DEFAULT_LANG']:  # Nothing to do
            return path
        else:  # Change lang and return
            return config['TRANSLATIONS_PATTERN'].format(path=p, ext=e, lang=lang)


def write_metadata(data, metadata_format=None, comment_wrap=False, site=None, compiler=None):
    """Write metadata.

    Recommended usage: pass `site`, `comment_wrap` (True, False, or a 2-tuple of start/end markers), and optionally `compiler`. Other options are for backwards compatibility.
    """
    # API compatibility
    if metadata_format is None and site is not None:
        metadata_format = site.config.get('METADATA_FORMAT', 'nikola').lower()
    if metadata_format is None:
        metadata_format = 'nikola'

    if site is None:
        import nikola.metadata_extractors
        metadata_extractors_by = nikola.metadata_extractors.default_metadata_extractors_by()
        nikola.metadata_extractors.load_defaults(site, metadata_extractors_by)
    else:
        metadata_extractors_by = site.metadata_extractors_by

    # Pelican is mapped to rest_docinfo, markdown_meta, or nikola.
    if metadata_format == 'pelican':
        if compiler and compiler.name == 'rest':
            metadata_format = 'rest_docinfo'
        elif compiler and compiler.name == 'markdown':
            metadata_format = 'markdown_meta'
        else:
            # Quiet fallback.
            metadata_format = 'nikola'

    default_meta = ('nikola', 'rest_docinfo', 'markdown_meta')
    extractor = metadata_extractors_by['name'].get(metadata_format)
    if extractor and extractor.supports_write:
        extractor.check_requirements()
        return extractor.write_metadata(data, comment_wrap)
    elif extractor and metadata_format not in default_meta:
        LOGGER.warning('Writing METADATA_FORMAT {} is not supported, using "nikola" format'.format(metadata_format))
    elif metadata_format not in default_meta:
        LOGGER.warning('Unknown METADATA_FORMAT {}, using "nikola" format'.format(metadata_format))

    if metadata_format == 'rest_docinfo':
        title = data['title']
        results = [
            '=' * len(title),
            title,
            '=' * len(title),
            ''
        ] + [':{0}: {1}'.format(k, v) for k, v in data.items() if v and k != 'title'] + ['']
        return '\n'.join(results)
    elif metadata_format == 'markdown_meta':
        results = ['{0}: {1}'.format(k, v) for k, v in data.items() if v] + ['', '']
        return '\n'.join(results)
    else:  # Nikola, default
        from nikola.metadata_extractors import DEFAULT_EXTRACTOR
        return DEFAULT_EXTRACTOR.write_metadata(data, comment_wrap)


def bool_from_meta(meta, key, fallback=False, blank=None):
    """Convert a boolean-ish meta value to a boolean."""
    value = meta.get(key)
    if isinstance(value, str):
        value_lowercase = value.lower().strip()
        if value_lowercase in {"true", "yes", "1"}:
            return True
        elif value_lowercase in {"false", "no", "0"}:
            return False
        elif not value_lowercase:
            return blank
    elif isinstance(value, int):
        return bool(value)
    elif value is None:
        return blank
    return fallback


def ask(query, default=None):
    """Ask a question."""
    if default:
        default_q = ' [{0}]'.format(default)
    else:
        default_q = ''
    inp = input("{query}{default_q}: ".format(query=query, default_q=default_q)).strip()
    if inp or default is None:
        return inp
    else:
        return default


def ask_yesno(query, default=None):
    """Ask a yes/no question."""
    if default is None:
        default_q = ' [y/n]'
    elif default is True:
        default_q = ' [Y/n]'
    elif default is False:
        default_q = ' [y/N]'
    inp = input("{query}{default_q} ".format(query=query, default_q=default_q)).strip()
    if inp:
        return inp.lower().startswith('y')
    elif default is not None:
        return default
    else:
        # Loop if no answer and no default.
        return ask_yesno(query, default)


class CommandWrapper(object):
    """Converts commands into functions."""

    def __init__(self, cmd, commands_object):
        self.cmd = cmd
        self.commands_object = commands_object

    def __call__(self, *args, **kwargs):
        if args or (not args and not kwargs):
            self.commands_object._run([self.cmd] + list(args))
        else:
            # Here's where the keyword magic would have to go
            self.commands_object._run_with_kw(self.cmd, *args, **kwargs)


class Commands(object):
    """Nikola Commands.

    Sample usage:
    >>> commands.check('-l')                     # doctest: +SKIP

    Or, if you know the internal argument names:
    >>> commands.check(list=True)                # doctest: +SKIP
    """

    def __init__(self, main, config, doitargs):
        """Take a main instance, work as wrapper for commands."""
        self._cmdnames = []
        self._main = main
        self._config = config
        self._doitargs = doitargs
        try:
            cmdict = self._doitargs['cmds'].to_dict()
        except AttributeError:  # not a doit PluginDict
            cmdict = self._doitargs['cmds']
        for k, v in cmdict.items():
            # cleanup: run is doit-only, init is useless in an existing site
            if k in ['run', 'init']:
                continue

            self._cmdnames.append(k)

            try:
                # nikola command: already instantiated (singleton)
                opt = v.get_options()
            except TypeError:
                # doit command: needs some help
                opt = v(config=self._config, **self._doitargs).get_options()
            nc = type(
                k,
                (CommandWrapper,),
                {
                    '__doc__': options2docstring(k, opt)
                })
            setattr(self, k, nc(k, self))

    def _run(self, cmd_args):
        self._main.run(cmd_args)

    def _run_with_kw(self, cmd, *a, **kw):
        # cyclic import hack
        from nikola.plugin_categories import Command
        try:
            cmd = self._doitargs['cmds'].get_plugin(cmd)
        except AttributeError:  # not a doit PluginDict
            cmd = self._doitargs['cmds'][cmd]
        try:
            opt = cmd.get_options()
        except TypeError:
            cmd = cmd(config=self._config, **self._doitargs)
            opt = cmd.get_options()

        options, _ = CmdParse(opt).parse([])
        options.update(kw)
        if isinstance(cmd, Command):
            cmd.execute(options=options, args=a)
        else:  # Doit command
            cmd.execute(options, a)

    def __repr__(self):
        """Return useful and verbose help."""
        return """\
<Nikola Commands>

    Sample usage:
    >>> commands.check('-l')

    Or, if you know the internal argument names:
    >>> commands.check(list=True)

Available commands: {0}.""".format(', '.join(self._cmdnames))


def options2docstring(name, options):
    """Translate options to a docstring."""
    result = ['Function wrapper for command %s' % name, 'arguments:']
    for opt in options:
        result.append('{0} type {1} default {2}'.format(opt.name, opt.type.__name__, opt.default))
    return '\n'.join(result)


class NikolaPygmentsHTML(BetterHtmlFormatter):
    """A Nikola-specific modification of Pygments' HtmlFormatter."""

    def __init__(self, anchor_ref=None, classes=None, **kwargs):
        """Initialize formatter."""
        if classes is None:
            classes = ['code', 'literal-block']
        if anchor_ref:
            kwargs['lineanchors'] = slugify(
                anchor_ref, lang=LocaleBorg().current_lang, force=True)
        self.nclasses = classes
        kwargs['cssclass'] = 'code'
        if not kwargs.get('linenos'):
            # Default to no line numbers (Issue #3426)
            kwargs['linenos'] = False
        if kwargs.get('linenos') not in {'table', 'inline', 'ol', False}:
            # Map invalid values to table
            kwargs['linenos'] = 'table'
        kwargs['anchorlinenos'] = kwargs['linenos'] == 'table'
        kwargs['nowrap'] = False
        super().__init__(**kwargs)

    def wrap(self, source, *args):
        """Wrap the ``source``, which is a generator yielding individual lines, in custom generators."""
        style = []
        if self.prestyles:
            style.append(self.prestyles)
        if self.noclasses:
            style.append('line-height: 125%')
        style = '; '.join(style)
        classes = ' '.join(self.nclasses)

        yield 0, ('<pre class="{0}"'.format(classes) + (style and ' style="{0}"'.format(style)) + '>')
        for tup in source:
            yield tup
        yield 0, '</pre>'


# For consistency, override the default formatter.
pygments.formatters._formatter_cache['HTML'] = NikolaPygmentsHTML
pygments.formatters._formatter_cache['html'] = NikolaPygmentsHTML
_original_find_formatter_class = pygments.formatters.find_formatter_class


def nikola_find_formatter_class(alias):
    """Nikola-specific version of find_formatter_class."""
    if "html" in alias.lower():
        return NikolaPygmentsHTML
    return _original_find_formatter_class(alias)


pygments.formatters.find_formatter_class = nikola_find_formatter_class


def get_displayed_page_number(i, num_pages, site):
    """Get page number to be displayed for entry `i`."""
    if not i:
        i = 0
    if site.config["INDEXES_STATIC"]:
        return i if i > 0 else num_pages
    else:
        return i + 1 if site.config["INDEXES_PAGES_MAIN"] else i


def adjust_name_for_index_path_list(path_list, i, displayed_i, lang, site, force_addition=False, extension=None):
    """Retrurn a path list for a given index page."""
    index_file = site.config["INDEX_FILE"]
    if i or force_addition:
        path_list = list(path_list)
        if force_addition and not i:
            i = 0
        if not extension:
            _, extension = os.path.splitext(index_file)
        if len(path_list) > 0 and path_list[-1] == '':
            path_list[-1] = index_file
        elif len(path_list) == 0 or not path_list[-1].endswith(extension):
            path_list.append(index_file)
        if site.config["PRETTY_URLS"] and site.config["INDEXES_PRETTY_PAGE_URL"](lang) and path_list[-1] == index_file:
            path_schema = site.config["INDEXES_PRETTY_PAGE_URL"](lang)
            if isinstance(path_schema, (bytes, str)):
                path_schema = [path_schema]
        else:
            path_schema = None
        if path_schema is not None:
            del path_list[-1]
            for entry in path_schema:
                path_list.append(entry.format(number=displayed_i, old_number=i, index_file=index_file))
        else:
            path_list[-1] = '{0}-{1}{2}'.format(os.path.splitext(path_list[-1])[0], i, extension)
    return path_list


def os_path_split(path):
    """Split a path."""
    result = []
    while True:
        previous_path = path
        path, tail = os.path.split(path)
        if path == previous_path and tail == '':
            result.insert(0, path)
            break
        result.insert(0, tail)
        if len(path) == 0:
            break
    return result


def adjust_name_for_index_path(name, i, displayed_i, lang, site, force_addition=False, extension=None):
    """Return file name for a given index file."""
    return os.path.join(*adjust_name_for_index_path_list(os_path_split(name), i, displayed_i, lang, site, force_addition, extension))


def adjust_name_for_index_link(name, i, displayed_i, lang, site, force_addition=False, extension=None):
    """Return link for a given index file."""
    link = adjust_name_for_index_path_list(name.split('/'), i, displayed_i, lang, site, force_addition, extension)
    if not extension == ".atom":
        if len(link) > 0 and link[-1] == site.config["INDEX_FILE"] and site.config["STRIP_INDEXES"]:
            link[-1] = ''
    return '/'.join(link)


def create_redirect(src, dst):
    """Create a redirection."""
    makedirs(os.path.dirname(src))
    with io.open(src, "w+", encoding="utf8") as fd:
        fd.write('<!DOCTYPE html>\n<head>\n<meta charset="utf-8">\n'
                 '<title>Redirecting...</title>\n<meta name="robots" '
                 'content="noindex">\n<meta http-equiv="refresh" content="0; '
                 'url={0}">\n</head>\n<body>\n<p>Page moved '
                 '<a href="{0}">here</a>.</p>\n</body>'.format(dst))


def colorize_str_from_base_color(string, base_color):
    """Find a perceptual similar color from a base color based on the hash of a string.

    Make up to 16 attempts (number of bytes returned by hashing) at picking a
    hue for our color at least 27 deg removed from the base color, leaving
    lightness and saturation untouched using HSLuv colorspace.
    """
    def hash_str(string, pos):
        return hashlib.md5(string.encode('utf-8')).digest()[pos]

    def degreediff(dega, degb):
        return min(abs(dega - degb), abs((degb - dega) + 360))

    if hsluv is None:
        req_missing(['hsluv'], 'Use color mixing (section colors)',
                    optional=True)
        return base_color
    h, s, l = hsluv.hex_to_hsluv(base_color)
    old_h = h
    idx = 0
    while degreediff(old_h, h) < 27 and idx < 16:
        h = 360.0 * (float(hash_str(string, idx)) / 255)
        idx += 1
    return hsluv.hsluv_to_hex((h, s, l))


def colorize_str(string: str, base_color: str, presets: dict):
    """Colorize a string by using a presets dict or generate one based on base_color."""
    if string in presets:
        return presets[string]
    return colorize_str_from_base_color(string, base_color)


def color_hsl_adjust_hex(hexstr, adjust_h=None, adjust_s=None, adjust_l=None):
    """Adjust a hex color using HSL arguments, adjustments in percentages 1.0 to -1.0. Returns a hex color."""
    h, s, l = hsluv.hex_to_hsluv(hexstr)

    if adjust_h:
        h = h + (adjust_h * 360.0)

    if adjust_s:
        s = s + (adjust_s * 100.0)

    if adjust_l:
        l = l + (adjust_l * 100.0)

    return hsluv.hsluv_to_hex((h, s, l))


def dns_sd(port, inet6):
    """Optimistically publish a HTTP service to the local network over DNS-SD.

    Works only on Linux/FreeBSD.  Requires the `avahi` and `dbus` modules (symlinks in virtualenvs)
    """
    try:
        import avahi
        import dbus
        inet = avahi.PROTO_INET6 if inet6 else avahi.PROTO_INET
        name = "{0}'s Nikola Server on {1}".format(os.getlogin(), socket.gethostname())
        bus = dbus.SystemBus()
        bus_server = dbus.Interface(bus.get_object(avahi.DBUS_NAME,
                                                   avahi.DBUS_PATH_SERVER),
                                    avahi.DBUS_INTERFACE_SERVER)
        bus_group = dbus.Interface(bus.get_object(avahi.DBUS_NAME,
                                                  bus_server.EntryGroupNew()),
                                   avahi.DBUS_INTERFACE_ENTRY_GROUP)
        bus_group.AddService(avahi.IF_UNSPEC, inet, dbus.UInt32(0),
                             name, '_http._tcp', '', '',
                             dbus.UInt16(port), '')
        bus_group.Commit()
        return bus_group  # remember to bus_group.Reset() to unpublish
    except Exception:
        return None


def clean_before_deployment(site):
    """Clean drafts and future posts before deployment."""
    undeployed_posts = []
    deploy_drafts = site.config.get('DEPLOY_DRAFTS', True)
    deploy_future = site.config.get('DEPLOY_FUTURE', False)
    if not (deploy_drafts and deploy_future):  # == !drafts || !future
        # Remove drafts and future posts
        out_dir = site.config['OUTPUT_FOLDER']
        site.scan_posts()
        for post in site.timeline:
            if (not deploy_drafts and post.is_draft) or (not deploy_future and post.publish_later):
                for lang in post.translated_to:
                    remove_file(os.path.join(out_dir, post.destination_path(lang)))
                    source_path = post.destination_path(lang, post.source_ext(True))
                    remove_file(os.path.join(out_dir, source_path))
                undeployed_posts.append(post)
    return undeployed_posts


def sort_posts(posts, *keys):
    """Sort posts by a given predicate. Helper function for templates.

    If a key starts with '-', it is sorted in descending order.

    Usage examples::

        sort_posts(timeline, 'title', 'date')
        sort_posts(timeline, 'author', '-section_name')
    """
    # We reverse the keys to get the usual ordering method: the first key
    # provided is the most important sorting predicate (first by 'title', then
    # by 'date' in the first example)
    for key in reversed(keys):
        if key.startswith('-'):
            key = key[1:]
            reverse = True
        else:
            reverse = False
        try:
            # An attribute (or method) of the Post object
            a = getattr(posts[0], key)
            if callable(a):
                keyfunc = operator.methodcaller(key)
            else:
                keyfunc = operator.attrgetter(key)
        except AttributeError:
            # Post metadata
            keyfunc = operator.methodcaller('meta', key)

        posts = sorted(posts, reverse=reverse, key=keyfunc)
    return posts


def smartjoin(join_char: str, string_or_iterable) -> str:
    """Join string_or_iterable with join_char if it is iterable; otherwise converts it to string.

    >>> smartjoin('; ', 'foo, bar')
    'foo, bar'
    >>> smartjoin('; ', ['foo', 'bar'])
    'foo; bar'
    >>> smartjoin(' to ', ['count', 42])
    'count to 42'
    """
    if isinstance(string_or_iterable, (str, bytes)):
        return string_or_iterable
    elif isinstance(string_or_iterable, Iterable):
        return join_char.join([str(e) for e in string_or_iterable])
    else:
        return str(string_or_iterable)


def _smartjoin_filter(string_or_iterable, join_char: str) -> str:
    """Join stuff smartly, with reversed arguments for Jinja2 filters."""
    # http://jinja.pocoo.org/docs/2.10/api/#custom-filters
    return smartjoin(join_char, string_or_iterable)


# Stolen from textwrap in Python 3.4.3.
def indent(text, prefix, predicate=None):
    """Add 'prefix' to the beginning of selected lines in 'text'.

    If 'predicate' is provided, 'prefix' will only be added to the lines
    where 'predicate(line)' is True. If 'predicate' is not provided,
    it will default to adding 'prefix' to all non-empty lines that do not
    consist solely of whitespace characters.
    """
    if predicate is None:
        def predicate(line):
            return line.strip()

    def prefixed_lines():
        for line in text.splitlines(True):
            yield (prefix + line if predicate(line) else line)
    return ''.join(prefixed_lines())


def load_data(path):
    """Given path to a file, load data from it."""
    ext = os.path.splitext(path)[-1]
    loader = None
    function = 'load'
    if ext in {'.yml', '.yaml'}:
        if YAML is None:
            req_missing(['ruamel.yaml'], 'use YAML data files')
            return {}
        loader = YAML(typ='safe')
        function = 'load'
    elif ext in {'.json', '.js'}:
        loader = json
    elif ext in {'.toml', '.tml'}:
        if toml is None:
            req_missing(['toml'], 'use TOML data files')
            return {}
        loader = toml
    if loader is None:
        return
    with io.open(path, 'r', encoding='utf-8-sig') as inf:
        return getattr(loader, function)(inf)


def rss_writer(rss_obj, output_path):
    """Write an RSS object to an xml file."""
    dst_dir = os.path.dirname(output_path)
    makedirs(dst_dir)
    with io.open(output_path, "w+", encoding="utf-8") as rss_file:
        data = rss_obj.to_xml(encoding='utf-8')
        if isinstance(data, bytes):
            data = data.decode('utf-8')
        rss_file.write(data)


def map_metadata(meta, key, config):
    """Map metadata from other platforms to Nikola names.

    This uses the METADATA_MAPPING and METADATA_VALUE_MAPPING settings (via ``config``) and modifies the dict in place.
    """
    for foreign, ours in config.get('METADATA_MAPPING', {}).get(key, {}).items():
        if foreign in meta:
            meta[ours] = meta[foreign]

    for meta_key, hook in config.get('METADATA_VALUE_MAPPING', {}).get(key, {}).items():
        if meta_key in meta:
            meta[meta_key] = hook(meta[meta_key])


def parselinenos(spec: str, total: int) -> List[int]:
    """Parse a line number spec.

    Example: "1,2,4-6" -> [0, 1, 3, 4, 5]
    """
    items = list()
    parts = spec.split(',')
    for part in parts:
        try:
            begend = part.strip().split('-')
            if ['', ''] == begend:
                raise ValueError
            elif len(begend) == 1:
                items.append(int(begend[0]) - 1)
            elif len(begend) == 2:
                start = int(begend[0] or 1)  # left half open (cf. -10)
                end = int(begend[1] or max(start, total))  # right half open (cf. 10-)
                if start > end:  # invalid range (cf. 10-1)
                    raise ValueError
                items.extend(range(start - 1, end))
            else:
                raise ValueError
        except Exception as exc:
            raise ValueError('invalid line number spec: %r' % spec) from exc
    return items


class ClassificationTranslationManager(object):
    """Keeps track of which classifications could be translated as which others.

    The internal structure is as follows:
    - per language, you have a map of classifications to maps
    - the inner map is a map from other languages to sets of classifications
      which are considered as translations
    """

    def __init__(self):
        self._data = defaultdict(dict)

    def add_translation(self, translation_map):
        """Add translation of one classification.

        ``translation_map`` must be a dictionary mapping languages to their
        translations of the added classification.
        """
        for lang, classification in translation_map.items():
            clmap = self._data[lang]
            cldata = clmap.get(classification)
            if cldata is None:
                cldata = defaultdict(set)
                clmap[classification] = cldata
            for other_lang, other_classification in translation_map.items():
                if other_lang != lang:
                    cldata[other_lang].add(other_classification)

    def get_translations(self, classification, lang):
        """Get a dict mapping other languages to (unsorted) lists of translated classifications."""
        clmap = self._data[lang]
        cldata = clmap.get(classification)
        if cldata is None:
            return {}
        else:
            return {other_lang: list(classifications) for other_lang, classifications in cldata.items()}

    def get_translations_as_list(self, classification, lang, classifications_per_language):
        """Get a list of pairs ``(other_lang, other_classification)`` which are translations of ``classification``.

        Avoid classifications not in ``classifications_per_language``.
        """
        clmap = self._data[lang]
        cldata = clmap.get(classification)
        if cldata is None:
            return []
        else:
            result = []
            for other_lang, classifications in cldata.items():
                for other_classification in classifications:
                    if other_classification in classifications_per_language[other_lang]:
                        result.append((other_lang, other_classification))
            return result

    def has_translations(self, classification, lang):
        """Return whether we know about the classification in that language.

        Note that this function returning ``True`` does not mean that
        ``get_translations`` returns a non-empty dict or that
        ``get_translations_as_list`` returns a non-empty list, but only
        that this classification was explicitly added with
        ``add_translation`` at some point.
        """
        return self._data[lang].get(classification) is not None

    def add_defaults(self, posts_per_classification_per_language):
        """Treat every classification as its own literal translation into every other language.

        ``posts_per_classification_per_language`` should be the first argument
        to ``Taxonomy.postprocess_posts_per_classification``.
        """
        # First collect all classifications from all languages
        all_classifications = set()
        for _, classifications in posts_per_classification_per_language.items():
            all_classifications.update(classifications.keys())
        # Next, add translation records for all of them
        for classification in all_classifications:
            record = {tlang: classification for tlang in posts_per_classification_per_language}
            self.add_translation(record)

    def read_from_config(self, site, basename, posts_per_classification_per_language, add_defaults_default):
        """Read translations from config.

        ``site`` should be the Nikola site object. Will consider
        the variables ``<basename>_TRANSLATIONS`` and
        ``<basename>_TRANSLATIONS_ADD_DEFAULTS``.

        ``posts_per_classification_per_language`` should be the first argument
        to ``Taxonomy.postprocess_posts_per_classification``, i.e. this function
        should be called from that function. ``add_defaults_default`` specifies
        what the default value for ``<basename>_TRANSLATIONS_ADD_DEFAULTS`` is.

        Also sends signal via blinker to allow interested plugins to add
        translations by themselves. The signal name used is
        ``<lower(basename)>_translations_config``, and the argument is a dict
        with entries ``translation_manager``, ``site`` and
        ``posts_per_classification_per_language``.
        """
        # Add translations
        for record in site.config.get('{}_TRANSLATIONS'.format(basename), []):
            self.add_translation(record)
        # Add default translations
        if site.config.get('{}_TRANSLATIONS_ADD_DEFAULTS'.format(basename), add_defaults_default):
            self.add_defaults(posts_per_classification_per_language)
        # Use blinker to inform interested parties (plugins) that they can add
        # translations themselves
        args = {'translation_manager': self, 'site': site,
                'posts_per_classification_per_language': posts_per_classification_per_language}
        signal('{}_translations_config'.format(basename.lower())).send(args)



# -*- encoding:utf-8 -*-
"""Autogenerated file, do not edit. Submit translations on Transifex."""

MESSAGES = {
    "%d min remaining to read": "%d دقیقه برای خواندن باقی مانده",
    "(active)": "(فعال)",
    "Also available in:": "همچنین قابل دسترس از:",
    "Archive": "آرشیو",
    "Atom feed": "",
    "Authors": "نویسنده‌ها",
    "Categories": "دسته‌ها",
    "Comments": "دیدگاه‌‌‌ها",
    "LANGUAGE": "فارسی",
    "Languages:": "زبان‌‌ها:",
    "More posts about %s": "ارسال‌های بیشتر دربارهٔ%s",
    "Newer posts": "ارسال‌های جدید‌تر",
    "Next post": "ارسال بعدی",
    "Next": "بعدی",
    "No posts found.": "هیچ پستی پیدا نشد.",
    "Nothing found.": "هیچ‌چیزی پیدا نشد.",
    "Older posts": "پست‌های قدیمی‌تر",
    "Original site": "سایت اصلی",
    "Posted:": "ارسال شده:",
    "Posts about %s": "ارسال‌ها دربارهٔ %s",
    "Posts by %s": "ارسال‌های %s",
    "Posts for year %s": "ارسال‌ها برای سال %s",
    "Posts for {month_day_year}": "ارسال برای {month_day_year}",
    "Posts for {month_year}": "ارسال برای {month_year}",
    "Previous post": "ارسال پیشین",
    "Previous": "قبلی",
    "Publication date": "تاریخ انتشار",
    "RSS feed": "خوراک",
    "Read in English": "به فارسی بخوانید",
    "Read more": "بیشتر بخوانید",
    "Skip to main content": "متن اصلی را نادیده بگیر",
    "Source": "منبع",
    "Subcategories:": "زیر بخش‌ها:",
    "Tags and Categories": "برچسب‌ها و دسته‌ها",
    "Tags": "برچسب‌ها",
    "Toggle navigation": "",
    "Uncategorized": "دسته‌بندی نشده",
    "Up": "",
    "Updates": "بروزرسانی‌ها",
    "Write your page here.": "من صفحه را این‌جا بنویسید. ",
    "Write your post here.": "متن پست‌تان را این‌جا بنویسید.",
    "old posts, page %d": "صفحهٔ ارسال‌های قدیمی %d",
    "page %d": "برگه %d",
    "updated": "",
}



# -*- coding: utf-8 -*-

# Copyright © 2012-2024 Roberto Alsina and others.

# Permission is hereby granted, free of charge, to any
# person obtaining a copy of this software and associated
# documentation files (the "Software"), to deal in the
# Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the
# Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice
# shall be included in all copies or substantial portions of
# the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY
# KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
# PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS
# OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR
# OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
# OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

"""Page compiler plugin for Markdown."""

import io
import json
import os
import threading

from nikola import shortcodes as sc
from nikola.plugin_categories import PageCompiler
from nikola.utils import makedirs, req_missing, write_metadata, LocaleBorg, map_metadata, NikolaPygmentsHTML

try:
    from markdown import Markdown
except ImportError:
    Markdown = None

# Override Pygments formatter for Markdown.
try:
    import markdown.extensions.codehilite
    markdown.extensions.codehilite.get_formatter_by_name = lambda _, **args: NikolaPygmentsHTML(**args)
except ImportError:
    pass


class ThreadLocalMarkdown(threading.local):
    """Convert Markdown to HTML using per-thread Markdown objects.

    See discussion in #2661.
    """

    def __init__(self, extensions, extension_configs):
        """Create a Markdown instance."""
        self.markdown = Markdown(extensions=extensions, extension_configs=extension_configs, output_format="html5")

    def convert(self, data):
        """Convert data to HTML and reset internal state."""
        result = self.markdown.convert(data)
        try:
            meta = {}
            for k in self.markdown.Meta:  # This reads everything as lists
                meta[k.lower()] = ','.join(self.markdown.Meta[k])
        except Exception:
            meta = {}
        self.markdown.reset()
        return result, meta


class CompileMarkdown(PageCompiler):
    """Compile Markdown into HTML."""

    name = "markdown"
    friendly_name = "Markdown"
    demote_headers = True
    site = None
    supports_metadata = False

    def set_site(self, site):
        """Set Nikola site."""
        super().set_site(site)
        self.config_dependencies = []
        extensions = []
        for plugin_info in self.get_compiler_extensions():
            self.config_dependencies.append(plugin_info.name)
            extensions.append(plugin_info.plugin_object)
            plugin_info.plugin_object.short_help = plugin_info.description

        site_extensions = self.site.config.get("MARKDOWN_EXTENSIONS")
        self.config_dependencies.append(str(sorted(site_extensions)))
        extensions.extend(site_extensions)

        site_extension_configs = self.site.config.get("MARKDOWN_EXTENSION_CONFIGS")
        if site_extension_configs:
            self.config_dependencies.append(json.dumps(site_extension_configs.values, sort_keys=True))

        if Markdown is not None:
            self.converters = {}
            for lang in self.site.config['TRANSLATIONS']:
                lang_extension_configs = site_extension_configs(lang) if site_extension_configs else {}
                self.converters[lang] = ThreadLocalMarkdown(extensions, lang_extension_configs)
        self.supports_metadata = 'markdown.extensions.meta' in extensions

    def compile_string(self, data, source_path=None, is_two_file=True, post=None, lang=None):
        """Compile Markdown into HTML strings."""
        if lang is None:
            lang = LocaleBorg().current_lang
        if Markdown is None:
            req_missing(['markdown'], 'build this site (compile Markdown)')
        if not is_two_file:
            _, data = self.split_metadata(data, post, lang)
        new_data, shortcodes = sc.extract_shortcodes(data)
        output, _ = self.converters[lang].convert(new_data)
        output, shortcode_deps = self.site.apply_shortcodes_uuid(output, shortcodes, filename=source_path, extra_context={'post': post})
        return output, shortcode_deps

    def compile(self, source, dest, is_two_file=True, post=None, lang=None):
        """Compile the source file into HTML and save as dest."""
        if Markdown is None:
            req_missing(['markdown'], 'build this site (compile Markdown)')
        makedirs(os.path.dirname(dest))
        with io.open(dest, "w+", encoding="utf-8") as out_file:
            with io.open(source, "r", encoding="utf-8-sig") as in_file:
                data = in_file.read()
            output, shortcode_deps = self.compile_string(data, source, is_two_file, post, lang)
            out_file.write(output)
        if post is None:
            if shortcode_deps:
                self.logger.error(
                    "Cannot save dependencies for post {0} (post unknown)",
                    source)
        else:
            post._depfile[dest] += shortcode_deps

    def create_post(self, path, **kw):
        """Create a new post."""
        content = kw.pop('content', None)
        onefile = kw.pop('onefile', False)
        # is_page is not used by create_post as of now.
        kw.pop('is_page', False)

        metadata = {}
        metadata.update(self.default_metadata)
        metadata.update(kw)
        makedirs(os.path.dirname(path))
        if not content.endswith('\n'):
            content += '\n'
        with io.open(path, "w+", encoding="utf-8") as fd:
            if onefile:
                fd.write(write_metadata(metadata, comment_wrap=True, site=self.site, compiler=self))
            fd.write(content)

    def read_metadata(self, post, lang=None):
        """Read the metadata from a post, and return a metadata dict."""
        lang = lang or self.site.config['DEFAULT_LANG']
        if not self.supports_metadata:
            return {}
        if Markdown is None:
            req_missing(['markdown'], 'build this site (compile Markdown)')
        if lang is None:
            lang = LocaleBorg().current_lang
        source = post.translated_source_path(lang)
        with io.open(source, 'r', encoding='utf-8-sig') as inf:
            # Note: markdown meta returns lowercase keys
            data = inf.read()
            # If the metadata starts with "---" it's actually YAML and
            # we should not let markdown parse it, because it will do
            # bad things like setting empty tags to "''"
            if data.startswith('---\n'):
                return {}
            _, meta = self.converters[lang].convert(data)
        # Map metadata from other platforms to names Nikola expects (Issue #2817)
        map_metadata(meta, 'markdown_metadata', self.site.config)
        return meta



# -*- encoding:utf-8 -*-
"""Autogenerated file, do not edit. Submit translations on Transifex."""

MESSAGES = {
    "%d min remaining to read": "पढ़ने के लिए %d मिनट शेष",
    "(active)": "(सक्रिय)",
    "Also available in:": "उपलब्ध भाषाएँ:",
    "Archive": "संग्रह",
    "Atom feed": "एटम फीड",
    "Authors": "लेखक",
    "Categories": "श्रेणियाँ",
    "Comments": "टिप्पणियाँ",
    "LANGUAGE": "हिन्दी",
    "Languages:": "भाषाएँ:",
    "More posts about %s": "%s के बारे में और पोस्ट",
    "Newer posts": "नई पोस्ट",
    "Next post": "अगला पोस्ट",
    "Next": "अगला",
    "No posts found.": "कोई पोस्ट नहीं मिला",
    "Nothing found.": "कुछ नहीं मिला",
    "Older posts": "पुरानी पोस्ट",
    "Original site": "मूल साइट",
    "Posted:": "पोस्टेड:",
    "Posts about %s": "%s के बारे में पोस्ट",
    "Posts by %s": "%s द्वारा पोस्ट",
    "Posts for year %s": "वर्ष  %s के पोस्ट",
    "Posts for {month_day_year}": "{month_day_year}  के पोस्ट",
    "Posts for {month_year}": "{month_year}  के पोस्ट",
    "Previous post": "पिछली पोस्ट",
    "Previous": "पिछला",
    "Publication date": "प्रकाशन तिथि",
    "RSS feed": "आर एस एस फ़ीड",
    "Read in English": "हिन्दी में पढ़िए",
    "Read more": "और पढ़िए",
    "Skip to main content": "मुख्य विषयवस्तु में जाएं",
    "Source": "स्रोत",
    "Subcategories:": "उपश्रेणी",
    "Tags and Categories": "टैग्स और श्रेणियाँ",
    "Tags": "टैग्स",
    "Toggle navigation": "टॉगल नेविगेशन",
    "Uncategorized": "अवर्गीकृत",
    "Up": "ऊपर",
    "Updates": "अपडेट्स",
    "Write your page here.": "अपना पेज यहाँ लिखें",
    "Write your post here.": "अपनी पोस्ट यहाँ लिखें",
    "old posts, page %d": "पुराने पोस्ट, पृष्‍ठ %d",
    "page %d": "पृष्‍ठ %d",
    "updated": "संशोधित",
}



# -*- coding: utf-8 -*-

# Copyright © 2012-2024 Roberto Alsina and others.

# Permission is hereby granted, free of charge, to any
# person obtaining a copy of this software and associated
# documentation files (the "Software"), to deal in the
# Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the
# Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice
# shall be included in all copies or substantial portions of
# the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY
# KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
# PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS
# OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR
# OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
# OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

"""reStructuredText compiler for Nikola."""

import io
import logging
import os

import docutils.core
import docutils.nodes
import docutils.transforms
import docutils.utils
import docutils.io
import docutils.readers.standalone
import docutils.writers.html5_polyglot
import docutils.parsers.rst.directives
from docutils.parsers.rst import roles

from nikola.nikola import LEGAL_VALUES
from nikola.metadata_extractors import MetaCondition
from nikola.plugin_categories import PageCompiler
from nikola.utils import (
    makedirs,
    write_metadata,
    LocaleBorg,
    map_metadata
)


class CompileRest(PageCompiler):
    """Compile reStructuredText into HTML."""

    name = "rest"
    friendly_name = "reStructuredText"
    demote_headers = True
    logger = None
    supports_metadata = True
    metadata_conditions = [(MetaCondition.config_bool, "USE_REST_DOCINFO_METADATA")]

    def read_metadata(self, post, lang=None):
        """Read the metadata from a post, and return a metadata dict."""
        if lang is None:
            lang = LocaleBorg().current_lang
        source_path = post.translated_source_path(lang)

        # Silence reST errors, some of which are due to a different
        # environment. Real issues will be reported while compiling.
        null_logger = logging.getLogger('NULL')
        null_logger.setLevel(1000)
        with io.open(source_path, 'r', encoding='utf-8-sig') as inf:
            data = inf.read()
            _, _, _, document = rst2html(data, logger=null_logger, source_path=source_path, transforms=self.site.rst_transforms)
        meta = {}
        if 'title' in document:
            meta['title'] = document['title']
        for docinfo in document.findall(docutils.nodes.docinfo):
            for element in docinfo.children:
                if element.tagname == 'field':  # custom fields (e.g. summary)
                    name_elem, body_elem = element.children
                    name = name_elem.astext()
                    value = body_elem.astext()
                elif element.tagname == 'authors':  # author list
                    name = element.tagname
                    value = [element.astext() for element in element.children]
                else:  # standard fields (e.g. address)
                    name = element.tagname
                    value = element.astext()
                name = name.lower()

                meta[name] = value

        # Put 'authors' meta field contents in 'author', too
        if 'authors' in meta and 'author' not in meta:
            meta['author'] = '; '.join(meta['authors'])

        # Map metadata from other platforms to names Nikola expects (Issue #2817)
        map_metadata(meta, 'rest_docinfo', self.site.config)
        return meta

    def compile_string(self, data, source_path=None, is_two_file=True, post=None, lang=None):
        """Compile reST into HTML strings."""
        # If errors occur, this will be added to the line number reported by
        # docutils so the line number matches the actual line number (off by
        # 7 with default metadata, could be more or less depending on the post).
        add_ln = 0
        if not is_two_file:
            m_data, data = self.split_metadata(data, post, lang)
            add_ln = len(m_data.splitlines()) + 1

        default_template_path = os.path.join(os.path.dirname(__file__), 'template.txt')
        settings_overrides = {
            'initial_header_level': 1,
            'record_dependencies': True,
            'stylesheet_path': None,
            'link_stylesheet': True,
            'syntax_highlight': 'short',
            # This path is not used by Nikola, but we need something to silence
            # warnings about it from reST.
            'math_output': 'mathjax /assets/js/mathjax.js',
            'template': default_template_path,
            'language_code': LEGAL_VALUES['DOCUTILS_LOCALES'].get(LocaleBorg().current_lang, 'en'),
            'doctitle_xform': self.site.config.get('USE_REST_DOCINFO_METADATA'),
            'file_insertion_enabled': self.site.config.get('REST_FILE_INSERTION_ENABLED'),
        }

        from nikola import shortcodes as sc
        new_data, shortcodes = sc.extract_shortcodes(data)
        if self.site.config.get('HIDE_REST_DOCINFO', False):
            self.site.rst_transforms.append(RemoveDocinfo)
        output, error_level, deps, _ = rst2html(
            new_data, settings_overrides=settings_overrides, logger=self.logger, source_path=source_path, l_add_ln=add_ln, transforms=self.site.rst_transforms)
        if not isinstance(output, str):
            # To prevent some weird bugs here or there.
            # Original issue: empty files.  `output` became a bytestring.
            output = output.decode('utf-8')

        output, shortcode_deps = self.site.apply_shortcodes_uuid(output, shortcodes, filename=source_path, extra_context={'post': post})
        return output, error_level, deps, shortcode_deps

    def compile(self, source, dest, is_two_file=True, post=None, lang=None):
        """Compile the source file into HTML and save as dest."""
        makedirs(os.path.dirname(dest))
        error_level = 100
        with io.open(dest, "w+", encoding="utf-8") as out_file:
            with io.open(source, "r", encoding="utf-8-sig") as in_file:
                data = in_file.read()
                output, error_level, deps, shortcode_deps = self.compile_string(data, source, is_two_file, post, lang)
                out_file.write(output)
            if post is None:
                if deps.list:
                    self.logger.error(
                        "Cannot save dependencies for post {0} (post unknown)",
                        source)
            else:
                post._depfile[dest] += deps.list
                post._depfile[dest] += shortcode_deps
        if error_level < 3:
            return True
        else:
            return False

    def create_post(self, path, **kw):
        """Create a new post."""
        content = kw.pop('content', None)
        onefile = kw.pop('onefile', False)
        # is_page is not used by create_post as of now.
        kw.pop('is_page', False)
        metadata = {}
        metadata.update(self.default_metadata)
        metadata.update(kw)
        makedirs(os.path.dirname(path))
        if not content.endswith('\n'):
            content += '\n'
        with io.open(path, "w+", encoding="utf-8") as fd:
            if onefile:
                fd.write(write_metadata(metadata, comment_wrap=False, site=self.site, compiler=self))
            fd.write(content)

    def set_site(self, site):
        """Set Nikola site."""
        super().set_site(site)
        self.config_dependencies = []
        for plugin_info in self.get_compiler_extensions():
            self.config_dependencies.append(plugin_info.name)
            plugin_info.plugin_object.short_help = plugin_info.description

        if not site.debug:
            self.logger.level = logging.WARNING


def get_observer(settings):
    """Return an observer for the docutils Reporter."""
    def observer(msg):
        """Report docutils/rest messages to a Nikola user.

        Error code mapping:

        +----------+----------+
        | docutils |  logging |
        +----------+----------+
        |    DEBUG |    DEBUG |
        |     INFO |     INFO |
        |  WARNING |  WARNING |
        |    ERROR |    ERROR |
        |   SEVERE | CRITICAL |
        +----------+----------+
        """
        errormap = {
            docutils.utils.Reporter.DEBUG_LEVEL: logging.DEBUG,
            docutils.utils.Reporter.INFO_LEVEL: logging.INFO,
            docutils.utils.Reporter.WARNING_LEVEL: logging.WARNING,
            docutils.utils.Reporter.ERROR_LEVEL: logging.ERROR,
            docutils.utils.Reporter.SEVERE_LEVEL: logging.CRITICAL
        }
        text = docutils.nodes.Element.astext(msg)
        line = msg['line'] + settings['add_ln'] if 'line' in msg else ''
        out = '[{source}{colon}{line}] {text}'.format(
            source=settings['source'], colon=(':' if line else ''),
            line=line, text=text)
        settings['logger'].log(errormap[msg['level']], out)

    return observer


class NikolaReader(docutils.readers.standalone.Reader):
    """Nikola-specific docutils reader."""

    config_section = 'nikola'

    def __init__(self, *args, **kwargs):
        """Initialize the reader."""
        self.transforms = kwargs.pop('transforms', [])
        self.logging_settings = kwargs.pop('nikola_logging_settings', {})
        docutils.readers.standalone.Reader.__init__(self, *args, **kwargs)

    def get_transforms(self):
        """Get docutils transforms."""
        return docutils.readers.standalone.Reader(self).get_transforms() + self.transforms

    def new_document(self):
        """Create and return a new empty document tree (root node)."""
        document = docutils.utils.new_document(self.source.source_path, self.settings)
        document.reporter.stream = False
        document.reporter.attach_observer(get_observer(self.logging_settings))
        return document


def shortcode_role(name, rawtext, text, lineno, inliner,
                   options={}, content=[]):
    """Return a shortcode role that passes through raw inline HTML."""
    return [docutils.nodes.raw('', text, format='html')], []


roles.register_canonical_role('raw-html', shortcode_role)
roles.register_canonical_role('html', shortcode_role)
roles.register_canonical_role('sc', shortcode_role)


def add_node(node, visit_function=None, depart_function=None):
    """Register a Docutils node class.

    This function is completely optional. It is a same concept as
    `Sphinx add_node function <https://www.sphinx-doc.org/en/master/extdev/appapi.html#sphinx.application.Sphinx.add_node>`_.

    For example::

        class Plugin(RestExtension):

            name = "rest_math"

            def set_site(self, site):
                self.site = site
                directives.register_directive('math', MathDirective)
                add_node(MathBlock, visit_Math, depart_Math)
                return super().set_site(site)

        class MathDirective(Directive):
            def run(self):
                node = MathBlock()
                return [node]

        class Math(docutils.nodes.Element): pass

        def visit_Math(self, node):
            self.body.append(self.starttag(node, 'math'))

        def depart_Math(self, node):
            self.body.append('</math>')

    For full example, you can refer to `Microdata plugin <https://plugins.getnikola.com/#microdata>`_
    """
    docutils.nodes._add_node_class_names([node.__name__])
    if visit_function:
        setattr(docutils.writers.html5_polyglot.HTMLTranslator, 'visit_' + node.__name__, visit_function)
    if depart_function:
        setattr(docutils.writers.html5_polyglot.HTMLTranslator, 'depart_' + node.__name__, depart_function)


# Output <code> for ``double backticks``. (Code and extra logic based on html4css1 translator)
def visit_literal(self, node):
    """Output <code> for double backticks."""
    # special case: "code" role
    classes = node.get('classes', [])
    if 'code' in classes:
        # filter 'code' from class arguments
        node['classes'] = [cls for cls in classes if cls != 'code']
        self.body.append(self.starttag(node, 'code', ''))
        return
    self.body.append(
        self.starttag(node, 'code', '', CLASS='docutils literal'))
    text = node.astext()
    for token in self.words_and_spaces.findall(text):
        if token.strip():
            # Protect text like "--an-option" and the regular expression
            # ``[+]?(\d+(\.\d*)?|\.\d+)`` from bad line wrapping
            if self.in_word_wrap_point.search(token):
                self.body.append('<span class="pre">%s</span>'
                                 % self.encode(token))
            else:
                self.body.append(self.encode(token))
        elif token in ('\n', ' '):
            # Allow breaks at whitespace:
            self.body.append(token)
        else:
            # Protect runs of multiple spaces; the last space can wrap:
            self.body.append('&nbsp;' * (len(token) - 1) + ' ')
    self.body.append('</code>')
    # Content already processed:
    raise docutils.nodes.SkipNode


setattr(docutils.writers.html5_polyglot.HTMLTranslator, 'visit_literal', visit_literal)


def rst2html(source, source_path=None, source_class=docutils.io.StringInput,
             destination_path=None, reader=None,
             parser=None, parser_name='restructuredtext', writer=None,
             writer_name='html5_polyglot', settings=None, settings_spec=None,
             settings_overrides=None, config_section='nikola',
             enable_exit_status=None, logger=None, l_add_ln=0, transforms=None):
    """Set up & run a ``Publisher``, and return a dictionary of document parts.

    Dictionary keys are the names of parts, and values are Unicode strings;
    encoding is up to the client.  For programmatic use with string I/O.

    For encoded string input, be sure to set the 'input_encoding' setting to
    the desired encoding.  Set it to 'unicode' for unencoded Unicode string
    input.  Here's how::

        publish_parts(..., settings_overrides={'input_encoding': 'unicode'})

    For a description of the parameters, see `publish_programmatically`.

    WARNING: `reader` should be None (or NikolaReader()) if you want Nikola to report
             reStructuredText syntax errors.
    """
    if reader is None:
        # For our custom logging, we have special needs and special settings we
        # specify here.
        # logger    a logger from Nikola
        # source   source filename (docutils gets a string)
        # add_ln   amount of metadata lines (see comment in CompileRest.compile above)
        reader = NikolaReader(transforms=transforms,
                              nikola_logging_settings={
                                  'logger': logger, 'source': source_path,
                                  'add_ln': l_add_ln
                              })

    pub = docutils.core.Publisher(reader, parser, writer, settings=settings,
                                  source_class=source_class,
                                  destination_class=docutils.io.StringOutput)
    pub.set_components(None, parser_name, writer_name)
    pub.process_programmatic_settings(
        settings_spec, settings_overrides, config_section)
    pub.set_source(source, None)
    pub.settings._nikola_source_path = source_path
    pub.set_destination(None, destination_path)
    pub.publish(enable_exit_status=enable_exit_status)

    return pub.writer.parts['docinfo'] + pub.writer.parts['fragment'], pub.document.reporter.max_level, pub.settings.record_dependencies, pub.document


# Alignment helpers for extensions
_align_options_base = ('left', 'center', 'right')


def _align_choice(argument):
    return docutils.parsers.rst.directives.choice(argument, _align_options_base + ("none", ""))


class RemoveDocinfo(docutils.transforms.Transform):
    """Remove docinfo nodes."""

    default_priority = 870

    def apply(self):
        """Remove docinfo nodes."""
        for node in self.document.traverse(docutils.nodes.docinfo):
            node.parent.remove(node)
